import os
from numbers import Number
from pathlib import Path
from typing import Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.font_manager import FontProperties

from insitupy.io.plots import save_and_show_figure


def volcano_plot(data,
                 logfoldchanges_column: str = 'logfoldchanges',
                 pval_column: str = 'neg_log10_pvals',
                 significance_threshold: Number = 0.05,
                 fold_change_threshold: Number = 1,
                 title: str = None,
                 adjust_labels: bool = True,
                 savepath: Union[str, os.PathLike, Path] = None,
                 save_only: bool = False,
                 dpi_save: int = 300,
                 label_top_n: int = 20,
                 figsize: Tuple[int, int] = (8, 6),
                 config_table=None
                 ):
    """
    Create a volcano plot from the DataFrame and label the top 20 most significant up and down-regulated genes.
    For the generation of the input data `insitupy.utils.deg.create_deg_dataframe` can be used

    Args:
        data (pd.DataFrame): DataFrame containing gene names, log fold changes, and p-values.
        logfoldchanges_column (str): Column name for log fold changes (default is 'logfoldchanges').
        pval_column (str): Column name for negative log10 p-values (default is 'neg_log10_pvals').
        significance_threshold (float): P-value threshold for significance (default is 0.05).
        fold_change_threshold (float): Log2 fold change threshold for up/down regulation (default is 1).
        title (str): Title of the plot (default is "Volcano Plot").
        adjust_labels (bool, optional): If True, adjusts the labels to avoid overlap. Default is False.
        savepath (Union[str, os.PathLike, Path], optional): Path to save the plot (default is None).
        save_only (bool): If True, only save the plot without displaying it (default is False).
        dpi_save (int): Dots per inch (DPI) for saving the plot (default is 300).
        label_top_n (int): Number of top up- and downregulated genes to label in the plot (default is 20).
        figsize (Tuple[int, int]): Size of the figure in inches (default is (8, 6)).

    Returns:
        None
    """
    if adjust_labels:
        try:
            from adjustText import adjust_text
        except ImportError:
            raise ImportError("The 'adjustText' module is required for label adjustment. Please install it with `pip install adjusttext` or select adjust_labels=False.")

    plt.figure(figsize=figsize)

    # Determine colors based on significance and fold change
    colors = []
    for index, row in data.iterrows():
        if row['pvals'] < significance_threshold:
            if row[logfoldchanges_column] > fold_change_threshold:
                colors.append('maroon')  # Up-regulated
            elif row[logfoldchanges_column] < -fold_change_threshold:
                colors.append('royalblue')  # Down-regulated
            else:
                colors.append('black')  # Not significant
        else:
            colors.append('black')  # Not significant

    # Scatter plot
    plt.scatter(data[logfoldchanges_column], data[pval_column],
                alpha=0.5, color=colors)

    # Add labels and title
    if title is not None:
        plt.title(title, fontsize=16)
    plt.xlabel('$\mathregular{Log_2}$ fold change', fontsize=14)
    plt.ylabel('$\mathregular{-Log_10}$ p-value', fontsize=14)

    # Add horizontal line for significance threshold
    plt.axhline(y=-np.log10(significance_threshold), color='black', linestyle='--')

    # Add vertical lines for fold change thresholds
    plt.axvline(x=fold_change_threshold, color='black', linestyle='--')
    plt.axvline(x=-fold_change_threshold, color='black', linestyle='--')

    # # Calculate mixed score and get top 20 up and down-regulated genes
    # volcano_data['mixed_score'] = -np.log10(volcano_data['pvals']) * volcano_data[logfoldchanges_column]

    top_up_genes = data[data[logfoldchanges_column] > fold_change_threshold].nlargest(label_top_n, 'scores')
    top_down_genes = data[data[logfoldchanges_column] < -fold_change_threshold].nsmallest(label_top_n, 'scores')

    # Combine top genes for annotation
    top_genes = pd.concat([top_up_genes, top_down_genes])

    # Adjust y-axis limits to provide space for text
    plt.ylim(0, plt.ylim()[1] * 1.1)  # Increase the upper limit of the y-axis to make space for the annotations

    # Annotate top genes
    texts = []
    for i, row in top_genes.iterrows():
        texts.append(plt.annotate(row['gene'],
                                   (row[logfoldchanges_column], row[pval_column]),
                                   fontsize=14,  # Increased font size
                                   alpha=0.75))

    if adjust_labels:
        # Adjust text to avoid overlap
        adjust_text(texts,
                    arrowprops=dict(arrowstyle='->', color='gray', lw=0.5),
                    max_move=None # this helped with some annotations remaining overlapping
                    )

    # Add labels to the top of the plot, outside the plot area
    plt.annotate('Target', xy=(1, 1.04), xycoords='axes fraction',
                xytext=(-65, 0), textcoords='offset points',
                ha='left', va='center', fontsize=14, color='black',
                arrowprops=dict(arrowstyle='->', color='black'))

    plt.annotate('Reference', xy=(0, 1.04), xycoords='axes fraction',
                xytext=(93, 0), textcoords='offset points',
                ha='right', va='center', fontsize=14, color='black',
                arrowprops=dict(arrowstyle='->', color='black'))

    if config_table is not None:
        # Create table data
        # Add table at the bottom of the plot
        table = plt.table(cellText=config_table.values,
                        colLabels=config_table.columns,
                        cellLoc='center',
                        colWidths=[.2,.4,.4],
                        loc='bottom',
                        bbox=[-0.12, -0.2-(0.1*(len(config_table)+1)), 1.12, 0.1*(len(config_table)+1)])

        # make first row and first column bold
        for (row, col), cell in table.get_celld().items():
            if (row == 0) | (col == 0):
                cell.set_text_props(fontproperties=FontProperties(weight='bold'))

        table.scale(xscale=2, yscale=1)
        # table.auto_set_font_size(True)
        # table.set_fontsize(12)

        # Adjust layout to make room for the table
        plt.subplots_adjust(left=0.2, bottom=0.1*(1+len(config_table)))

    # save and show figure
    save_and_show_figure(savepath=savepath, fig=plt.gcf(), save_only=save_only, dpi_save=dpi_save)
    plt.show()


from typing import List, Literal, Optional, Tuple, Union

import numpy as np
from anndata import AnnData

from insitupy._core._checks import _check_assignment
from insitupy._core.insitudata import InSituData
from insitupy.utils.utils import convert_to_list


# Define the function
def _check_string_in_assignment(entry, string_to_check):
    split_list = entry.split(" & ")
    return string_to_check in split_list

def _check_list_in_assignment(entry, list_to_check):
    split_list = set(entry.split(" & "))
    return not set(list_to_check).isdisjoint(split_list) # check if the sets are not disjoint - meaning at least one element is found in both

def _select_data_for_dge(
    data: InSituData,
    annotation_tuple: Optional[Union[Tuple[str, str], Tuple[str, List[str]]]] = None,
    cell_type_tuple: Optional[Tuple[str, str]] = None,
    region_tuple: Optional[Tuple[str, str]] = None,
    force_assignment: bool = False,
    verbose: bool = False
    ) -> AnnData:

    # extract anndata object
    adata = data.cells.matrix.copy()

    ### REGIONS
    if region_tuple is not None:
        # assign region
        _check_assignment(data=data,
                          key=region_tuple[0],
                          force_assignment=force_assignment,
                          modality="regions",
                          verbose=verbose
                          )

        # select only one region
        s = adata.obsm["regions"][region_tuple[0]]
        region_mask = s.apply(_check_string_in_assignment, string_to_check=region_tuple[1])
        if not np.any(region_mask):
            raise ValueError(f"Region '{region_tuple[1]}' not found in key '{region_tuple[0]}'.")

        if verbose:
            print(f"Restrict analysis to region '{region_tuple[1]}' from key '{region_tuple[0]}'.", flush=True)
        adata = adata[region_mask].copy()

    ### ANNOTATIONS
    if annotation_tuple is not None:
        # check if the annotations need to be assigned first
        _check_assignment(data=data,
                          key=annotation_tuple[0],
                          force_assignment=force_assignment,
                          modality="annotations",
                          verbose=verbose
                          )

        # create mask for filtering
        s = adata.obsm["annotations"][annotation_tuple[0]]
        if isinstance(annotation_tuple[1], str):
            annot_mask = s.apply(_check_string_in_assignment, string_to_check=annotation_tuple[1])
        elif isinstance(annotation_tuple[1], list):
            annot_mask = s.apply(_check_list_in_assignment, list_to_check=annotation_tuple[1])
        else:
            raise ValueError(f'Unknown type for second element of `annotation_tuple`: {type(annotation_tuple[1])}. Must be list or str.')

        if not np.any(annot_mask):
            raise ValueError(f"annotation_name '{annotation_tuple[1]}' not found under annotation_key '{annotation_tuple[0]}'.")

        # do filtering
        if verbose:
            print(f"Restrict analysis to annotation '{annotation_tuple[1]}' from key '{annotation_tuple[0]}'.", flush=True)
        adata = adata[annot_mask].copy()

    ### CELL TYPES
    if cell_type_tuple is not None:
        # create mask for filtering
        if isinstance(cell_type_tuple[1], str):
            cell_type_mask = adata.obs[cell_type_tuple[0]] == cell_type_tuple[1]
        elif isinstance(cell_type_tuple[1], list):
            cell_type_mask = adata.obs[cell_type_tuple[0]].isin(cell_type_tuple[1])
        else:
            raise ValueError(f'Unknown type for second element of `cell_type_tuple`: {type(cell_type_tuple[1])}. Must be list or str.')

        if not np.any(cell_type_mask):
            raise ValueError(f"Cell type '{cell_type_tuple[1]}' not found in .obs column '{cell_type_tuple[0]}'.")

        if verbose:
            print(f"Restrict analysis to cell type '{cell_type_tuple[1]}' from .obs column '{cell_type_tuple[0]}'.", flush=True)
        adata = adata[cell_type_mask].copy()

    return adata


def _substitution_func(
    row,
    annotation_key,
    annotation_name,
    check_for_duplicate_assignment_of_reference,
    reference_name=None,
    ignore_duplicate_assignments=False
    ):
    elem = row[annotation_key]
    # check_reference = False
    # if reference_name is not None:
    #     check_reference = True
    try:
        split_name = elem.split(" & ")
        if annotation_name in split_name:
            if check_for_duplicate_assignment_of_reference:
                if reference_name in split_name:
                    if not ignore_duplicate_assignments:
                        raise ValueError(f"Cell '{row.name}' was found to belong to both the annotation and the reference. To ignore this and use only the annotation assignment, use `ignore_duplicate_assignments=True`. Assignment that was found is: {elem}")
                    else:
                        return annotation_name
                else:
                    return annotation_name
            else:
                return annotation_name
        else:
            # TODO: Here also the reference name must be kept - does it really? Only if both annotation and reference are in the same dataset
            return elem
    except AttributeError:
        return elem


from warnings import warn


### DEPRECATED FUNCTIONS
def read_all(self, *args, **kwargs):
    warn("`read_all` is deprecated. Use `load_all` instead.", DeprecationWarning, stacklevel=2)

def read_annotations(self, *args, **kwargs):
    warn("`read_annotations` is deprecated. Use `load_annotations` instead.", DeprecationWarning, stacklevel=2)

def read_regions(self, *args, **kwargs):
    warn("`read_regions` is deprecated. Use `load_regions` instead.", DeprecationWarning, stacklevel=2)

def read_cells(self, *args, **kwargs):
    warn("`read_cells` is deprecated. Use `load_cells` instead.", DeprecationWarning, stacklevel=2)

def read_images(self, *args, **kwargs):
    warn("`read_images` is deprecated. Use `load_images` instead.", DeprecationWarning, stacklevel=2)

def read_transcripts(self, *args, **kwargs):
    warn("`read_transcripts` is deprecated. Use `load_transcripts` instead.", DeprecationWarning, stacklevel=2)

def read_xenium(self, *args, **kwargs):
    warn("`read_xenium` is deprecated. Use `read(mode='xenium')` instead.", DeprecationWarning, stacklevel=2)


from copy import deepcopy


class DeepCopyMixin:
    def copy(self):
        '''
        Function to generate a deep copy of the current object.
        '''
        return deepcopy(self)

class GetMixin:
    def get(self, key):
        '''
        Function to retrieve and return an attribute of the current object.
        '''
        return getattr(self, key)

    def __getitem__(self, key):
        '''
        Function to retrieve and return an attribute of the current object.
        '''
        return getattr(self, key)



from typing import Optional, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import cm
from matplotlib.colors import ListedColormap
from pandas.api.types import is_bool_dtype, is_numeric_dtype

from insitupy._constants import (DEFAULT_CATEGORICAL_CMAP,
                                 DEFAULT_CONTINUOUS_CMAP)
from insitupy.palettes import CustomPalettes


def create_cmap_mapping(data, cmap: Union[str, ListedColormap] = None):

    if cmap is None:
        pal = CustomPalettes()
        cmap = pal.tab20_mod
    try:
        unique_categories = data.cat.categories # in case of categorical pandas series
    except AttributeError:
        try:
            unique_categories = data.categories # in case of numpy categories
        except AttributeError:
            data = np.array(data)
            try:
                unique_categories = np.sort(data[~data.isna()].unique())
            except AttributeError:
                try:
                    unique_categories = np.sort(np.unique(data[~np.isnan(data)]))
                except TypeError:
                    #unique_categories = np.sort(np.unique(data))
                    # Convert all elements to strings before sorting
                    unique_categories = np.sort(np.unique(data.astype(str)))

    # get colormap if necessary
    if not isinstance(cmap, ListedColormap):
        cmap = plt.get_cmap(cmap)

    len_colormap = cmap.N
    category_to_rgba = {category: cmap(i % len_colormap) for i, category in enumerate(unique_categories)}
    return category_to_rgba


def categorical_data_to_rgba(data,
                             cmap: Union[str, ListedColormap],
                             return_mapping: bool = False,
                             nan_val: tuple = (1,1,1,0)
                             ):

    # len_colormap = cmap.N
    # category_to_rgba = {category: cmap(i % len_colormap) for i, category in enumerate(unique_categories)}

    if not isinstance(cmap, dict):
        category_to_rgba = create_cmap_mapping(data, cmap)
    else:
        category_to_rgba = cmap

    if nan_val is not None:
        # add key for nan
        category_to_rgba[str(np.nan)] = nan_val

    res = np.array([category_to_rgba[str(category)] for category in data])

    if return_mapping:
        return res, category_to_rgba
    else:
        return res


def _determine_climits(
    color_values,
    upper_climit_pct,
    lower_climit = None,
    force_above_zero = False
    ) -> list:

    if lower_climit is None:
        lower_climit = color_values.min()

    if force_above_zero:
        color_values_for_calc = color_values[color_values > 0]
    else:
        color_values_for_calc = color_values
    try:
        upper_climit = np.percentile(color_values_for_calc, upper_climit_pct)
    except IndexError:
        # if there were no color values above zero, a IndexError appears
        upper_climit = 0

    climits = [lower_climit, upper_climit]

    return climits


def continuous_data_to_rgba(
    data,
    cmap: Union[str, ListedColormap],
    upper_climit_pct: int = 99,
    lower_climit: Optional[int] = None,
    clip = False,
    nan_val: tuple = (1,1,1,0),
    return_mapping: bool = False
    ):
    if np.any(pd.isna(data)):
        contains_nans = True
        # Convert the numpy array to a pandas Series
        value_series = pd.Series(data)

        # check where there are nas
        isna_mask = value_series.isna()
        notna_mask = ~isna_mask

        # get only values without NaNs
        notna_values = value_series[notna_mask].values
    else:
        notna_values = data
        contains_nans = False

    # get colormap if necessary
    if not isinstance(cmap, ListedColormap):
        cmap = plt.get_cmap(cmap)

    if lower_climit is None:
        lower_climit = np.min(notna_values)

    climits = _determine_climits(color_values=notna_values, upper_climit_pct=upper_climit_pct, lower_climit=lower_climit)

    norm = mpl.colors.Normalize(vmin=climits[0], vmax=climits[1], clip=clip)
    scalarMap = cm.ScalarMappable(norm=norm, cmap=cmap)
    rgba_data = scalarMap.to_rgba(notna_values)

    if contains_nans:
        result_values = pd.Series(index=value_series.index, dtype="object")
        tuple_list = [tuple(elem) for elem in rgba_data]
        result_values.loc[notna_mask] = tuple_list
        result_values.loc[isna_mask] = [nan_val] * isna_mask.sum()
        result_rgba = np.array(result_values.tolist())
    else:
        result_rgba = rgba_data

    if return_mapping:
        return result_rgba, scalarMap
    else:
        return result_rgba


def _data_to_rgba(
    data: np.ndarray,
    continuous_cmap: Union[str, ListedColormap] = DEFAULT_CONTINUOUS_CMAP,
    categorical_cmap: Union[str, ListedColormap] = None,
    upper_climit_pct: int = 99,
    #return_all: bool = False,
    nan_val: tuple = (1,1,1,0),
    ):
    if not is_numeric_dtype(data) or is_bool_dtype(data):
        if is_bool_dtype(data):
            # in case of boolean data it is important to convert it to object type
            data = data.astype('object')
            # also pandas NAs have to be substituted with numpy NaNs
            data = np.array([elem if pd.notna(elem) else np.nan for elem in data], dtype='object')
        if categorical_cmap is None:
            # pal = CustomPalettes()
            # categorical_cmap = pal.tab20_mod
            categorical_cmap = DEFAULT_CATEGORICAL_CMAP
        rgba_list, mapping = categorical_data_to_rgba(data=data, cmap=categorical_cmap,
                                        return_mapping=True,
                                        nan_val=nan_val)
        cmap = categorical_cmap
    else:
        rgba_list, mapping = continuous_data_to_rgba(data=data, cmap=continuous_cmap,
                                       upper_climit_pct=upper_climit_pct,
                                       return_mapping=True)
        cmap = continuous_cmap

    return rgba_list, mapping, cmap


import math
import os
from datetime import datetime
from typing import Optional, Tuple, Union
from uuid import uuid4
from warnings import warn

import dask.dataframe as dd
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numpy import ndarray
from pandas.api.types import is_numeric_dtype, is_string_dtype
from parse import datetime
from shapely import LineString, Point, Polygon, affinity

from insitupy._constants import (XENIUM_HEX_TO_INT_CONV_DICT,
                                 XENIUM_INT_TO_HEX_CONV_DICT)


class textformat:
    '''
    Helper class to format printed text.
    e.g. print(color.RED + color.BOLD + 'Hello, World!' + color.END)
    '''
    # colors and formats
    # PURPLE = '\033[95m'
    # CYAN = '\033[96m'
    # DARKCYAN = '\033[36m'
    # BLUE = '\033[94m'
    # GREEN = '\033[92m'
    # YELLOW = '\033[93m'
    # RED = '\033[91m'
    # BOLD = '\033[1m'
    # UNDERLINE = '\033[4m'
    # END = '\033[0m'

    ResetAll = "\033[0m"

    Bold       = "\033[1m"
    Dim        = "\033[2m"
    Underlined = "\033[4m"
    Blink      = "\033[5m"
    Reverse    = "\033[7m"
    Hidden     = "\033[8m"

    ResetBold       = "\033[21m"
    ResetDim        = "\033[22m"
    ResetUnderlined = "\033[24m"
    ResetBlink      = "\033[25m"
    ResetReverse    = "\033[27m"
    ResetHidden     = "\033[28m"

    Default      = "\033[39m"
    Black        = "\033[30m"
    Red          = "\033[31m"
    Green        = "\033[32m"
    Yellow       = "\033[33m"
    Blue         = "\033[34m"
    Magenta      = "\033[35m"
    Cyan         = "\033[36m"
    LightGray    = "\033[37m"
    DarkGray     = "\033[90m"
    LightRed     = "\033[91m"
    LightGreen   = "\033[92m"
    LightYellow  = "\033[93m"
    LightBlue    = "\033[94m"
    Purple       = "\033[95m"
    LightCyan    = "\033[96m"
    White        = "\033[97m"

    BackgroundDefault      = "\033[49m"
    BackgroundBlack        = "\033[40m"
    BackgroundRed          = "\033[41m"
    BackgroundGreen        = "\033[42m"
    BackgroundYellow       = "\033[43m"
    BackgroundBlue         = "\033[44m"
    BackgroundMagenta      = "\033[45m"
    BackgroundCyan         = "\033[46m"
    BackgroundLightGray    = "\033[47m"
    BackgroundDarkGray     = "\033[100m"
    BackgroundLightRed     = "\033[101m"
    BackgroundLightGreen   = "\033[102m"
    BackgroundLightYellow  = "\033[103m"
    BackgroundLightBlue    = "\033[104m"
    BackgroundLightMagenta = "\033[105m"
    BackgroundLightCyan    = "\033[106m"
    BackgroundWhite        = "\033[107m"

    # signs
    TSIGN = "\u251c"
    LSIGN = "\u2514"
    HLINE = "\u2500"
    RARROWHEAD = "\u27A4"
    TICK = "\u2714"
    CIRCLE_EMPTY = "\u25EF"
    CIRCLE_FILLED = "\u2B24"
    CIRCLE_HALF = "\u25D0"
    CIRCLE_THREEQUARTER = "\u25D5"
    CIRCLE_ONEQUARTER = "\u25D4"

    # spacer
    SPACER = "    "

def create_ansi_color_code_from_rgb(rgb_color):
    # Create the ANSI escape code
    ansi_escape_code = f'\033[38;2;{rgb_color[0]};{rgb_color[1]};{rgb_color[2]}m'
    return ansi_escape_code

def remove_last_line_from_csv(filename):
    with open(filename) as myFile:
        lines = myFile.readlines()
        last_line = lines[len(lines)-1]
        lines[len(lines)-1] = last_line.rstrip()
    with open(filename, 'w') as myFile:
        myFile.writelines(lines)

def decode_robust(s, encoding="utf-8"):
    try:
        return s.decode(encoding)
    except (UnicodeDecodeError, AttributeError):
        return s

def decode_robust_series(s, encoding="utf-8"):
    '''
    Function to decode a pandas series in a robust fashion with different checks.
    This circumvents the return of NaNs and makes a decision in case of different errors.
    '''
    if is_numeric_dtype(s):
        return s
    if is_string_dtype(s):
        return s
    try:
        decoded = s.str.decode(encoding)
        if decoded.isna().all():
            return decoded
        elif decoded.isna().any():
            namask = decoded.isna()
            decoded[namask] = s[namask]
        return decoded
    except (UnicodeDecodeError, AttributeError):
        return s

def convert_to_list(elem):
    '''
    Return element to list if it is not a list already.
    '''
    return [elem] if (isinstance(elem, str) or isinstance(elem, os.PathLike) or isinstance(elem, int)) else list(elem)

def nested_dict_numpy_to_list(dictionary):
    for key, value in dictionary.items():
        if isinstance(value, ndarray):
            dictionary[key] = value.tolist()
        elif isinstance(value, dict):
            nested_dict_numpy_to_list(value)

def get_nrows_maxcols(n_keys, max_cols):
    '''
    Determine optimal number of rows and columns for `plt.subplot` based on
    number of keys ['n_keys'] and maximum number of columns [`max_cols`].

    Returns: `n_plots`, `n_rows`, `max_cols`
    '''

    #n_plots = len(keys)
    if n_keys > max_cols:
        n_rows = math.ceil(n_keys / max_cols)
    else:
        n_rows = 1
        max_cols = n_keys

    return n_keys, n_rows, max_cols

def remove_empty_subplots(axes, nplots, nrows, ncols):
    if nplots > 1:
        # check if there are empty plots remaining
        i = nplots
        while i < nrows * ncols:
            # remove empty plots
            axes[i].set_axis_off()
            i+=1

def check_list(List, list_to_compare):
    '''
    Compare two lists and return the elements that are in both lists.
    If not all elements are in both lists give message telling which are not.
    '''

    not_in = []
    List = [l if l in list_to_compare else not_in.append(l) for l in List]

    # remove None values
    List = [elem for elem in List if elem is not None]

    if len(not_in) > 0:
        print("Following elements not found: {}".format(", ".join(not_in)), flush=True)

    return List

def _generate_time_based_uid():
    time_str = datetime.now().strftime("%y%m%d-%H%M%S%f")
    short_uid = str(uuid4()).split("-")[0]
    uid = f"{time_str}-{short_uid}"
    return uid

def convert_int_to_xenium_hex(value, dataset_suffix=1, final_length=8):
    """Convert integers into Xenium-style hexadecimal representation.
    Described here: https://www.10xgenomics.com/support/software/xenium-onboard-analysis/latest/analysis/xoa-output-zarr#cellID

    Args:
        value (_type_): _description_
        dataset_suffix (int, optional): _description_. Defaults to 1.
        final_length (int, optional): _description_. Defaults to 8.

    Returns:
        _type_: _description_
    """
    # generate hexadecimal representation
    hex_repr = hex(value)[2:]

    # convert normal hex to xenium-modified hex
    hex_repr = "".join([str(XENIUM_INT_TO_HEX_CONV_DICT[elem]) for elem in hex_repr])

    # add a to the beginning to fill to final length
    hex_repr = hex_repr.rjust(final_length, 'a')

    # add dataset suffix
    hex_repr += f"-{dataset_suffix}"

    return hex_repr

def convert_xenium_hex_to_int(hex_repr):
    """Convert Xenium-style hexadecimal representation into integers.
    Described here: https://www.10xgenomics.com/support/software/xenium-onboard-analysis/latest/analysis/xoa-output-zarr#cellID


    Args:
        hex_repr (_type_): _description_

    Returns:
        _type_: _description_
    """
    # remove dataset suffix
    hex_repr_split = hex_repr.split("-")

    # try to extract a dataset suffix
    try:
        dataset_suffix = int(hex_repr_split[1])
    except IndexError:
        dataset_suffix = None

    # extract the hex repr
    hex_repr = hex_repr_split[0]

    # remove leading a
    hex_repr = hex_repr.lstrip("a")

    # convert xenium-modified hex to normal hex
    hex_repr = "".join([str(XENIUM_HEX_TO_INT_CONV_DICT[elem]) for elem in hex_repr])

    # generate decimal representation
    dec = int(hex_repr, 16)

    return dec, dataset_suffix


def create_ellipse_from_bbox(corner_coords):
    """
    Create an ellipse from a bounding box defined by its corner coordinates.

    Parameters:
    corner_coords (list of tuples): A list containing the coordinates of the four corners
                                     of the bounding box in the format [(x1, y1), (x2, y2),
                                     (x3, y3), (x4, y4)].

    Returns:
    shapely.geometry.Polygon: A Shapely polygon representing the ellipse.
    """
    # Unpack the corner coordinates
    x_coords = [coord[0] for coord in corner_coords]
    y_coords = [coord[1] for coord in corner_coords]

    # Calculate the bounding box's min and max coordinates
    x_min = min(x_coords)
    x_max = max(x_coords)
    y_min = min(y_coords)
    y_max = max(y_coords)

    # Calculate the center of the bounding box
    center_x = (x_min + x_max) / 2
    center_y = (y_min + y_max) / 2

    # Calculate the semi-major and semi-minor axes
    semi_major = (x_max - x_min) / 2
    semi_minor = (y_max - y_min) / 2

    # Create a unit circle centered at the origin
    unit_circle = Point(0, 0).buffer(1)  # Create a unit circle

    # Scale the unit circle to create the ellipse
    ellipse = affinity.scale(unit_circle, xfact=semi_major, yfact=semi_minor)

    # Translate the ellipse to the center of the bounding box
    ellipse = affinity.translate(ellipse, xoff=center_x, yoff=center_y)

    return ellipse

def convert_napari_shape_to_polygon_or_line(napari_shape_data, shape_type):
    """
    Convert shape data from Napari format to a Shapely polygon or line.

    This function takes shape data in the format used by Napari and converts it
    into a Shapely geometry object, which can be either a polygon, an ellipse,
    or a line string, depending on the specified shape type.

    Parameters:
    napari_shape_data (numpy.ndarray): An array of shape data, where each row
                                        represents a point (y, x) in the Napari
                                        coordinate system.
    shape_type (str): A string indicating the type of shape to create.
                      Accepted values are:
                      - "polygon": Converts the shape data to a Shapely Polygon.
                      - "ellipse": Converts the shape data to a Shapely Polygon
                                   representing an ellipse based on the bounding box.
                      - "path": Converts the shape data to a Shapely LineString.

    Returns:
    shapely.geometry.Polygon or shapely.geometry.LineString: A Shapely geometry
    object representing the converted shape.

    Raises:
    TypeError: If the provided shape_type is not one of the accepted values.
    """
    if (shape_type == "polygon") or (shape_type == "rectangle"):
        result = Polygon(np.stack([napari_shape_data[:, 1], napari_shape_data[:, 0]], axis=1))
    elif shape_type == "ellipse":
        result = create_ellipse_from_bbox(np.flip(napari_shape_data, axis=1))
    elif shape_type == "path":
        result = LineString(np.flip(napari_shape_data, axis=1))
    else:
        raise TypeError(f"Shape has an unknown type: {shape_type}")

    return result

def exclude_index(array, exclude_index):
    """
    Exclude the element at the specified index from the array.

    Args:
        array (np.ndarray): The input NumPy array.
        exclude_index (int): The index of the element to exclude.

    Returns:
        np.ndarray: A new array with the element at exclude_index excluded.
    """
    return np.concatenate((array[:exclude_index], array[exclude_index+1:]))

def _crop_transcripts(
    transcript_df: Union[pd.DataFrame, dd.DataFrame],
    shape: Optional[Polygon] = None,
    xlim: Optional[Tuple[int, int]] = None,
    ylim: Optional[Tuple[int, int]] = None,
    verbose: bool = True
    ):

    if shape is not None:
        if xlim is not None and ylim is not None:
            if verbose:
                warn("Both xlim/ylim and shape are provided. Shape will be used for cropping.")

        try:
            points = gpd.points_from_xy(x=transcript_df.loc[:, ("coordinates", "x")].values,
                                        y=transcript_df.loc[:, ("coordinates", "y")].values)
            warn("Filtering transcripts based on a shape may take longer if transcripts are stored as pandas dataframe instead of dask dataframe.")
            grouped_df = True
        except KeyError:
            try:
                import dask_geopandas as dask_gpd
            except ImportError:
                warn("Filtering transcripts based on a shape may take longer if `dask_geopandas` is not installed.")

                # load the dataframe into memory to generate points
                print("Load transcript dataframe into memory...")
                transcript_df = transcript_df.compute()
                # generate points without dask_geopandas
                points = gpd.points_from_xy(x=transcript_df.loc[:, "x_location"].values,
                                            y=transcript_df.loc[:, "y_location"].values)
            else:
                # generate points with dask_geopandas
                points = dask_gpd.points_from_xy(df=transcript_df, x="x_location", y="y_location")
            grouped_df = False

        # create mask
        #mask = shape.contains(points)
        mask = points.within(shape)

        # get minimum x and y values
        minx, miny, _, _ = shape.bounds

    else:
        if xlim is None or ylim is None:
            raise ValueError("Either both xlim and ylim must be provided, or shape must be provided.")

        try:
            # infer mask for selection
            xmask = (transcript_df["coordinates", "x"] >= xlim[0]) & (transcript_df["coordinates", "x"] <= xlim[1])
            ymask = (transcript_df["coordinates", "y"] >= ylim[0]) & (transcript_df["coordinates", "y"] <= ylim[1])
            grouped_df = True
        except KeyError:
            xmask = (transcript_df["x_location"] >= xlim[0]) & (transcript_df["x_location"] <= xlim[1])
            ymask = (transcript_df["y_location"] >= ylim[0]) & (transcript_df["y_location"] <= ylim[1])
            grouped_df = False

        # create filtering mask
        mask = xmask & ymask

        # get minimum x and y for shifting the coordinates after cropping
        minx = xlim[0]
        miny = ylim[0]

    # select
    transcript_df = transcript_df.loc[mask, :].copy()

    if grouped_df:
        # move origin again to 0 by subtracting the lower limits from the coordinates
        transcript_df["coordinates", "x"] -= minx
        transcript_df["coordinates", "y"] -= miny
    else:
        # move origin again to 0 by subtracting the lower limits from the coordinates
        transcript_df["x_location"] -= minx
        transcript_df["y_location"] -= miny

    return transcript_df



import functools as ft
import gc
import json
import os
import shutil
from datetime import datetime
from numbers import Number
from os.path import abspath
from pathlib import Path
from typing import List, Literal, Optional, Tuple, Union
from uuid import uuid4
from warnings import catch_warnings, filterwarnings, warn

import anndata
import dask.dataframe as dd
import geopandas as gpd
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
from anndata._core.anndata import AnnData
from dask_image.imread import imread
from geopandas import GeoDataFrame
from parse import *
from scipy.sparse import issparse
from shapely import Point, Polygon
from shapely.affinity import scale as scale_func
from tqdm import tqdm

import insitupy._core.config as config
from insitupy import WITH_NAPARI, __version__
from insitupy._constants import ISPY_METADATA_FILE, LOAD_FUNCS, REGIONS_SYMBOL
from insitupy._core._checks import _check_assignment
from insitupy._core._save import _save_images
from insitupy._core._xenium import (_read_binned_expression,
                                    _read_boundaries_from_xenium,
                                    _read_matrix_from_xenium,
                                    _restructure_transcripts_dataframe)
from insitupy._exceptions import UnknownOptionError
from insitupy._warnings import NoProjectLoadWarning
from insitupy.images import ImageRegistration, deconvolve_he, resize_image
from insitupy.images.utils import _get_contrast_limits
from insitupy.io.files import read_json, write_dict_to_json
from insitupy.io.io import (read_baysor_cells, read_baysor_transcripts,
                            read_celldata, read_shapesdata)
from insitupy.io.plots import save_and_show_figure
from insitupy.plotting import volcano_plot
from insitupy.utils.dge import create_deg_dataframe
from insitupy.utils.preprocessing import (normalize_and_transform_anndata,
                                          reduce_dimensions_anndata)
from insitupy.utils.utils import (_crop_transcripts, convert_to_list,
                                  get_nrows_maxcols)

from .._constants import CACHE, ISPY_METADATA_FILE, MODALITIES
from .._exceptions import (InSituDataMissingObject,
                           InSituDataRepeatedCropError, ModalityNotFoundError,
                           NotOneElementError, WrongNapariLayerTypeError)
from ..images.utils import create_img_pyramid
from ..io.files import check_overwrite_and_remove_if_true, read_json
from ..plotting import expr_along_obs_val
from ..utils.utils import (convert_napari_shape_to_polygon_or_line,
                           convert_to_list)
from ..utils.utils import textformat as tf
from ._layers import _create_points_layer
from ._save import (_save_alt, _save_annotations, _save_cells, _save_images,
                    _save_regions, _save_transcripts)
from .dataclasses import AnnotationsData, CellData, ImageData, RegionsData

# optional packages that are not always installed
if WITH_NAPARI:
    import napari
    from napari.layers import Layer, Points, Shapes

    #from napari.layers.shapes.shapes import Shapes
    from ._layers import _add_geometries_as_layer
    from ._widgets import _initialize_widgets, add_new_geometries_widget


class InSituData:
    #TODO: Docstring of InSituData

    # import deprecated functions
    from ._deprecated import (read_all, read_annotations, read_cells,
                              read_images, read_regions, read_transcripts)

    def __init__(self,
                 path: Union[str, os.PathLike, Path] = None,
                 metadata: dict = None,
                 slide_id: str = None,
                 sample_id: str = None,
                 from_insitudata: bool = None,
                 ):
        """
        """
        # metadata
        self._path = Path(path)
        self._metadata = metadata
        self._slide_id = slide_id
        self._sample_id = sample_id
        self._from_insitudata = from_insitudata

        # modalities
        self._images = None
        self._cells = None
        self._alt = None
        self._annotations = None
        self._transcripts = None
        self._regions = None

        # other
        self._viewer = None
        self._quicksave_dir = None

    def __repr__(self):
        if self._metadata is None:
            method = "unknown"
        else:
            try:
                method = self._metadata["method"]
            except KeyError:
                method = "unknown"

        if self._path is not None:
            self._path = self._path.resolve()

        # check if all modalities are empty
        is_empty = np.all([elem is None for elem in [self._images, self._cells, self._alt, self._annotations, self._transcripts, self._regions]])

        # if is_empty:
        #     repr = f"{tf.Bold+tf.Red}InSituData{tf.ResetAll}\nEmpty"
        # else:
        repr = (
            f"{tf.Bold+tf.Red}InSituData{tf.ResetAll}\n"
            f"{tf.Bold}Method:{tf.ResetAll}\t\t{method}\n"
            f"{tf.Bold}Slide ID:{tf.ResetAll}\t{self._slide_id}\n"
            f"{tf.Bold}Sample ID:{tf.ResetAll}\t{self._sample_id}\n"
            f"{tf.Bold}Path:{tf.ResetAll}\t\t{self._path}\n"
        )

        if self._metadata is not None:
            if "metadata_file" in self._metadata:
                mfile = self._metadata["metadata_file"]
            else:
                mfile = None
        else:
            mfile = None

        repr += f"{tf.Bold}Metadata file:{tf.ResetAll}\t{mfile}"

        if is_empty:
            repr += "\n\nNo modalities loaded."
        else:
            if self._images is not None:
                images_repr = self._images.__repr__()
                repr = (
                    repr + f"\n{tf.SPACER+tf.RARROWHEAD} " + images_repr.replace("\n", f"\n{tf.SPACER}   ")
                )

            if self._cells is not None:
                cells_repr = self._cells.__repr__()
                repr = (
                    repr + f"\n{tf.SPACER+tf.RARROWHEAD+tf.Green+tf.Bold} cells{tf.ResetAll}\n{tf.SPACER}   " + cells_repr.replace("\n", f"\n{tf.SPACER}   ")
                )

            if self._transcripts is not None:
                trans_repr = f"DataFrame with shape {self._transcripts.shape[0]} x {self._transcripts.shape[1]}"

                repr = (
                    repr + f"\n{tf.SPACER+tf.RARROWHEAD+tf.Purple+tf.Bold} transcripts{tf.ResetAll}\n{tf.SPACER}   " + trans_repr
                )

            if self._annotations is not None:
                annot_repr = self._annotations.__repr__()
                repr = (
                    repr + f"\n{tf.SPACER+tf.RARROWHEAD} " + annot_repr.replace("\n", f"\n{tf.SPACER}   ")
                )

            if self._regions is not None:
                region_repr = self._regions.__repr__()
                repr = (
                    repr + f"\n{tf.SPACER+tf.RARROWHEAD} " + region_repr.replace("\n", f"\n{tf.SPACER}   ")
                )

            if self._alt is not None:
                cells_repr = self._alt.__repr__()
                altseg_keys = self._alt.keys()
                repr = (
                    #repr + f"\n{tf.SPACER+tf.RARROWHEAD+tf.Green+tf.Bold} alt{tf.ResetAll}\n{tf.SPACER}   " + cells_repr.replace("\n", f"\n{tf.SPACER}   ")
                    repr + f"\n{tf.SPACER+tf.RARROWHEAD+tf.Green+tf.Bold} alt{tf.ResetAll}\n"
                    f"{tf.SPACER}   Alternative CellData objects with following keys: {','.join(altseg_keys)}"
                )
        return repr


    @property
    def path(self):
        """Return save path of the InSituData object.
        Returns:
            str: Save path.
        """
        return self._path

    @property
    def metadata(self):
        """Return metadata of the InSituData object.
        Returns:
            dict: Metadata.
        """
        return self._metadata

    @metadata.setter
    def metadata(self, metadata: dict):
        self._metadata = metadata

    @property
    def slide_id(self):
        """Return slide id of the InSituData object.
        Returns:
            str: Slide id.
        """
        return self._slide_id

    @property
    def sample_id(self):
        """Return sample id of the InSituData object.
        Returns:
            str: Sample id.
        """
        return self._sample_id

    @property
    def from_insitudata(self):
        return self._from_insitudata

    @property
    def images(self):
        """Return images of the InSituData object.
        Returns:
            insitupy._core.dataclasses.ImageData: Images.
        """
        return self._images

    @images.setter
    def images(self, images: ImageData):
        self._images = images

    @images.deleter
    def images(self):
        self._images = None

    @property
    def cells(self):
        """Return cell data of the InSituData object.
        Returns:
            insitupy._core.dataclasses.CellData: Cell data.
        """
        return self._cells

    @cells.setter
    def cells(self, value: CellData):
        self._cells = value

    @cells.deleter
    def cells(self):
        self._cells = None

    @property
    def transcripts(self):
        """Return transcripts of the InSituData object.
        Returns:
            pd.DataFrame: Transcripts.
        """
        return self._transcripts

    @transcripts.setter
    def transcripts(self, value: pd.DataFrame):
        self._transcripts = value

    @transcripts.deleter
    def transcripts(self):
        self._transcripts = None

    @property
    def viewer(self):
        """Return viewer of the InSituData object.
        """
        return self._viewer

    @viewer.setter
    def viewer(self, value):
        self._viewer = value

    @viewer.deleter
    def viewer(self):
        self._viewer = None

    @property
    def annotations(self):
        """Return annotations of the InSituData object.
        Returns:
            insitupy._core.dataclasses.AnnotationsData: Annotations.
        """
        return self._annotations

    @annotations.setter
    def annotations(self, value: AnnotationsData):
        self._annotations = value

    @annotations.deleter
    def annotations(self):
        self._annotations = None

    @property
    def alt(self):
        """Return alternative cell data of the InSituData object.
        Returns:
            dict: A dictionary containing `insitupy._core.dataclasses.CellData` objects..
        """
        return self._alt

    @alt.deleter
    def alt(self):
        self._alt = None

    @property
    def regions(self):
        """Return regions of the InSituData object.
        Returns:
            insitupy._core.dataclasses.RegionsData: Regions.
        """
        return self._regions

    @regions.setter
    def regions(self, value: RegionsData):
        self._regions = value

    @regions.deleter
    def regions(self):
        self._regions = None

    def assign_geometries(self,
                          geometry_type: Literal["annotations", "regions"],
                          keys: Union[str, Literal["all"]] = "all",
                          add_masks: bool = False,
                          add_to_obs: bool = False,
                          overwrite: bool = True,
                          alt_layer: str = None
                          ):
        '''
        Function to assign geometries (annotations or regions) to the anndata object in
        InSituData.cells.matrix. Assignment information is added to the DataFrame in `.obs`.
        '''
        # assert that prerequisites are met
        try:
            geom_attr = getattr(self, geometry_type)
        except AttributeError:
            raise ModalityNotFoundError(modality=geometry_type)

        if alt_layer is None:
            if self._cells is not None:
                cell_attr = self._cells
                name = ".cells"
            else:
                raise ModalityNotFoundError("cells")
        else:
            #TODO
            if self._alt is not None:
                cell_attr = self._alt[alt_layer]
                name = f".alt[{alt_layer}]"
            else:
                raise ModalityNotFoundError(f"alt[{alt_layer}]")

        if keys == "all":
            keys = geom_attr.metadata.keys()

        # make sure annotation keys are a list
        keys = convert_to_list(keys)

        # convert coordinates into shapely Point objects
        x = cell_attr.matrix.obsm["spatial"][:, 0]
        y = cell_attr.matrix.obsm["spatial"][:, 1]
        cells = gpd.points_from_xy(x, y)

        # iterate through annotation keys
        for key in keys:
            print(f"Assigning key '{key}'...")
            # extract pandas dataframe of current key
            geom_df = geom_attr[key]

            # make sure the geom names do not contain any ampersand string (' % '),
            # since this would interfere with the downstream analysis
            if geom_df["name"].str.contains(' & ').any():
                raise ValueError(
                    f"The {geometry_type} with key '{key}' contains names with the ampersand string ' & '. "
                    f"This is not allowed as it would interfere with downstream analysis."
                    )

            # get unique list of annotation names
            geom_names = geom_df.name.unique()

            # initiate dataframe as dictionary
            data = {}

            # iterate through names
            for n in geom_names:
                polygons = geom_df[geom_df["name"] == n]["geometry"].tolist()
                #scales = geom_df[geom_df["name"] == n]["scale"].tolist()

                # in_poly = []
                # for poly, scale in zip(polygons, scales):
                #     # scale the polygon
                #     poly = scale_func(poly, xfact=scale[0], yfact=scale[1], origin=(0,0))

                #     # check if which of the points are inside the current annotation polygon
                #     in_poly.append(poly.contains(cells))

                in_poly = [poly.contains(cells) for poly in polygons]

                # check if points were in any of the polygons
                in_poly_res = np.array(in_poly).any(axis=0)

                # collect results
                data[n] = in_poly_res

            # convert into pandas dataframe
            data = pd.DataFrame(data)
            data.index = cell_attr.matrix.obs_names

            # transform data into one column
            column_to_add = [" & ".join(geom_names[row.values]) if np.any(row.values) else "unassigned" for _, row in data.iterrows()]

            if add_to_obs:
                # create annotation from annotation masks
                col_name = f"{geometry_type}-{key}"
                data[col_name] = column_to_add

                if col_name in cell_attr.matrix.obs:
                    if overwrite:
                        cell_attr.matrix.obs.drop(col_name, axis=1, inplace=True)
                        print(f'Existing column "{col_name}" is overwritten.', flush=True)
                        add = True
                    else:
                        warn(f'Column "{col_name}" exists already in `{name}.matrix.obs`. Assignment of key "{key}" was skipped. To force assignment, select `overwrite=True`.')
                        add = False

                if add:
                    if add_masks:
                        cell_attr.matrix.obs = pd.merge(left=cell_attr.matrix.obs, right=data, left_index=True, right_index=True)
                    else:
                        cell_attr.matrix.obs = pd.merge(left=cell_attr.matrix.obs, right=data.iloc[:, -1], left_index=True, right_index=True)

                    # save that the current key was analyzed
                    geom_attr.metadata[key]["analyzed"] = tf.TICK
            else:
                # add to obsm
                obsm_keys = cell_attr.matrix.obsm.keys()
                if geometry_type not in obsm_keys:
                    # add empty pandas dataframe with obs_names as index
                    cell_attr.matrix.obsm[geometry_type] = pd.DataFrame(index=cell_attr.matrix.obs_names)

                cell_attr.matrix.obsm[geometry_type][key] = column_to_add

                # save that the current key was analyzed
                geom_attr.metadata[key]["analyzed"] = tf.TICK

                print(f"Added results to `{name}.matrix.obsm['{geometry_type}']", flush=True)


    def assign_annotations(
        self,
        keys: Union[str, Literal["all"]] = "all",
        add_masks: bool = False,
        overwrite: bool = True
    ):
        self.assign_geometries(
            geometry_type="annotations",
            keys=keys,
            add_masks=add_masks,
            overwrite=overwrite
        )
        if self._alt is not None:
            for key in self.alt.keys():
                self.assign_geometries(
                    geometry_type="annotations",
                    keys=keys,
                    add_masks=add_masks,
                    overwrite=overwrite,
                    alt_layer=key
                )

    def assign_regions(
        self,
        keys: Union[str, Literal["all"]] = "all",
        add_masks: bool = False,
        overwrite: bool = True
    ):
        self.assign_geometries(
            geometry_type="regions",
            keys=keys,
            add_masks=add_masks,
            overwrite=overwrite
        )
        if self._alt is not None:
            for key in self.alt.keys():
                self.assign_geometries(
                    geometry_type="regions",
                    keys=keys,
                    add_masks=add_masks,
                    overwrite=overwrite,
                    alt_layer=key
                )

    def copy(self):
        '''
        Function to generate a deep copy of the InSituData object.
        '''
        from copy import deepcopy
        had_viewer = False
        if self._viewer is not None:
            had_viewer = True

            # make copy of viewer to add it later again
            viewer_copy = self._viewer.copy()
            # remove viewer because there is otherwise a error during deepcopy
            self.viewer = None

        # make copy
        self_copy = deepcopy(self)

        # add viewer again to original object if necessary
        if had_viewer:
            self._viewer = viewer_copy

        return self_copy

    def crop(self,
             region_tuple: Optional[Tuple[str, str]] = None,
             xlim: Optional[Tuple[int, int]] = None,
             ylim: Optional[Tuple[int, int]] = None,
             inplace: bool = False,
             verbose: bool = False
            ):
        """
        Crop the data based on the provided parameters.

        Args:
            region_tuple (Optional[Tuple[str, str]]): A tuple specifying the region to crop.
            xlim (Optional[Tuple[int, int]]): The x-axis limits for cropping.
            ylim (Optional[Tuple[int, int]]): The y-axis limits for cropping.
            inplace (bool): If True, modify the data in place. Otherwise, return a new cropped data.

        Raises:
            ValueError: If none of region_tuple, layer_name, or xlim/ylim are provided.
        """
        # check if the changes are supposed to be made in place or not
        if inplace:
            _self = self
        else:
            _self = self.copy()

        # if layer_name is None and region_tuple is None and (xlim is None or ylim is None):
        #     raise ValueError("At least one of shape_layer, region_tuple, or xlim/ylim must be provided.")
        if region_tuple is None:
            if xlim is None or ylim is None:
                raise ValueError("If shape is None, both xlim and ylim must not be None.")

            # make sure there are no negative values in the limits
            xlim = tuple(np.clip(xlim, a_min=0, a_max=None))
            ylim = tuple(np.clip(ylim, a_min=0, a_max=None))
            shape = None
        else:
            # extract regions dataframe
            region_key = region_tuple[0]
            region_name = region_tuple[1]
            region_df = self._regions[region_key]

            # extract geometry
            shape = region_df[region_df["name"] == region_name]["geometry"].item()
            #use_shape = True

            # extract x and y limits from the geometry
            minx, miny, maxx, maxy = shape.bounds # (minx, miny, maxx, maxy)
            xlim = (minx, maxx)
            ylim = (miny, maxy)

        try:
            # if the object was previously cropped, check if the current window is identical with the previous one
            if np.all([elem in _self.metadata["method_params"].keys() for elem in ["cropping_xlim", "cropping_ylim"]]):
                # test whether the limits are identical
                if (xlim == _self.metadata["method_params"]["cropping_xlim"]) & (ylim == _self.metadata["method_params"]["cropping_ylim"]):
                    raise InSituDataRepeatedCropError(xlim, ylim)
        except TypeError:
            pass

        if _self.cells is not None:
            _self.cells.crop(
                shape=shape,
                xlim=xlim, ylim=ylim,
                inplace=True, verbose=False
            )

        if _self.alt is not None:
            alt = _self.alt
            for k, alt_cells in alt.items():
                alt_cells.crop(
                    shape=shape,
                    xlim=xlim, ylim=ylim, inplace=True,
                    verbose=verbose
                )

        if _self.transcripts is not None:
            _self.transcripts = _crop_transcripts(
                transcript_df=_self.transcripts,
                shape=shape,
                xlim=xlim, ylim=ylim, verbose=verbose
            )

        if self._images is not None:
            _self.images.crop(xlim=xlim, ylim=ylim)

        if self._annotations is not None:

            _self.annotations.crop(
                shape=shape,
                xlim=tuple([elem for elem in xlim]),
                ylim=tuple([elem for elem in ylim]),
                verbose=verbose
                )

        if self._regions is not None:
            _self.regions.crop(
                shape=shape,
                xlim=tuple([elem for elem in xlim]),
                ylim=tuple([elem for elem in ylim]),
                verbose=verbose
            )

        if _self.metadata is not None:
            # add information about cropping to metadata
            if "cropping_history" not in _self.metadata:
                _self.metadata["cropping_history"] = {}
                _self.metadata["cropping_history"]["xlim"] = []
                _self.metadata["cropping_history"]["ylim"] = []
            _self.metadata["cropping_history"]["xlim"].append(tuple([int(elem) for elem in xlim]))
            _self.metadata["cropping_history"]["ylim"].append(tuple([int(elem) for elem in ylim]))

            # add new uid to uid history
            _self.metadata["uids"].append(str(uuid4()))

            # empty current data and data history entry in metadata
            _self.metadata["data"] = {}
            for k in _self.metadata["history"].keys():
                if k != "alt":
                    _self.metadata["history"][k] = []
                else:
                    empty_alt_hist_dict = {k: [] for k in _self.metadata["history"]["alt"].keys()}
                    _self.metadata["history"]["alt"] = empty_alt_hist_dict

        if inplace:
            if self._viewer is not None:
                del _self.viewer # delete viewer
        else:
            return _self


    def hvg(self,
            hvg_batch_key: Optional[str] = None,
            hvg_flavor: Literal["seurat", "cell_ranger", "seurat_v3"] = 'seurat',
            hvg_n_top_genes: Optional[int] = None,
            verbose: bool = True
            ) -> None:
        """
        Calculate highly variable genes (HVGs) using specified flavor and parameters.

        Args:
            hvg_batch_key (str, optional):
                Batch key for computing HVGs separately for each batch. Default is None, indicating all samples are considered.
            hvg_flavor (Literal["seurat", "cell_ranger", "seurat_v3"], optional):
                Flavor of the HVG computation method. Choose between "seurat", "cell_ranger", or "seurat_v3".
                Default is 'seurat'.
            hvg_n_top_genes (int, optional):
                Number of top highly variable genes to identify. Mandatory if `hvg_flavor` is set to "seurat_v3".
                Default is None.
            verbose (bool, optional):
                If True, print progress messages during HVG computation. Default is True.

        Raises:
            ValueError: If `hvg_n_top_genes` is not specified for "seurat_v3" flavor or if an invalid `hvg_flavor` is provided.

        Returns:
            None: This method modifies the input matrix in place, identifying highly variable genes based on the specified
                flavor and parameters. It does not return any value.
        """

        if hvg_flavor in ["seurat", "cell_ranger"]:
            hvg_layer = None
        elif hvg_flavor == "seurat_v3":
            hvg_layer = "counts" # seurat v3 method expects counts data

            # n top genes must be specified for this method
            if hvg_n_top_genes is None:
                raise ValueError(f"HVG computation: For flavor {hvg_flavor} `hvg_n_top_genes` is mandatory")
        else:
            raise ValueError(f'Unknown value for `hvg_flavor`: {hvg_flavor}. Possible values: {["seurat", "cell_ranger", "seurat_v3"]}')

        if hvg_batch_key is None:
            print("Calculate highly-variable genes across all samples using {} flavor...".format(hvg_flavor)) if verbose else None
        else:
            print("Calculate highly-variable genes per batch key {} using {} flavor...".format(hvg_batch_key, hvg_flavor)) if verbose else None

        sc.pp.highly_variable_genes(self._cells.matrix, batch_key=hvg_batch_key, flavor=hvg_flavor, layer=hvg_layer, n_top_genes=hvg_n_top_genes)


    def normalize_and_transform(self,
                transformation_method: Literal["log1p", "sqrt"] = "log1p",
                target_sum: int = 250,
                normalize_alt: bool = True,
                verbose: bool = True
                ) -> None:
        """
        Normalize the data using either log1p or square root transformation.

        Args:
            transformation_method (Literal["log1p", "sqrt"], optional):
                The method used for data transformation. Choose between "log1p" for logarithmic transformation
                and "sqrt" for square root transformation. Default is "log1p".
            normalize_alt (bool, optional):
                If True, `.alt` modalities are also normalized, if available.
            verbose (bool, optional):
                If True, print progress messages during normalization. Default is True.

        Raises:
            ValueError: If `transformation_method` is not one of ["log1p", "sqrt"].

        Returns:
            None: This method modifies the input matrix in place, normalizing the data based on the specified method.
                It does not return any value.
        """
        if self._cells is not None:
            cells = self._cells
        else:
            raise ModalityNotFoundError(modality="cells")

        normalize_and_transform_anndata(
            adata=cells.matrix,
            transformation_method=transformation_method,
            target_sum=target_sum,
            verbose=verbose)

        if self._alt is not None:
            alt = self._alt
            print("Found `.alt` modality.")
            for k, cells in alt.items():
                print(f"\tNormalizing {k}...")
                normalize_and_transform_anndata(adata=cells.matrix, transformation_method=transformation_method, verbose=verbose)

    def add_alt(self,
                celldata_to_add: CellData,
                key_to_add: str
                ) -> None:
        # check if the current self has already an alt object and add a empty one if not
        #alt_attr_name = "alt"
        #try:
        #    alt_attr = getattr(self, alt_attr_name)
        #except AttributeError:
        #    setattr(self, alt_attr_name, {})
        #    alt_attr = getattr(self, alt_attr_name)

        if self._alt is None:
            self._alt = dict()

        # add the celldata to the given key
        self._alt[key_to_add] = celldata_to_add

    def add_baysor(self,
                   path: Union[str, os.PathLike, Path],
                   read_transcripts: bool = False,
                   key_to_add: str = "baysor",
                   pixel_size: Number = 1 # the pixel size is usually 1 since baysor runs on the µm coordinates
                   ):

        # # convert to pathlib path
        path = Path(path)

        # read baysor data
        celldata = read_baysor_cells(baysor_output=path, pixel_size=pixel_size)

        # add celldata to alt attribute
        self.add_alt(celldata_to_add=celldata, key_to_add=key_to_add)

        if read_transcripts:
            #trans_attr_name = "transcripts"
            if self._transcripts is None:
                print("No transcript layer found. Addition of Baysor transcript data is skipped.", flush=True)
            else:
                trans_attr = self._transcripts
                # read baysor transcripts
                baysor_results = read_baysor_transcripts(baysor_output=path)
                baysor_results = baysor_results[["cell"]]

                # merge transcripts with existing transcripts
                baysor_results.columns = pd.MultiIndex.from_tuples([("cell_id", key_to_add)])
                trans_attr = pd.merge(left=trans_attr,
                                    right=baysor_results,
                                    left_index=True,
                                    right_index=True
                                    )

                # add resulting dataframe to InSituData
                self._transcripts = trans_attr


    def plot_dimred(self, save: Optional[str] = None):
        '''
        Read dimensionality reduction plots.
        '''
        # construct paths
        analysis_path = self._path / "analysis"
        umap_file = analysis_path / "umap" / "gene_expression_2_components" / "projection.csv"
        pca_file = analysis_path / "pca" / "gene_expression_10_components" / "projection.csv"
        cluster_file = analysis_path / "clustering" / "gene_expression_graphclust" / "clusters.csv"


        # read data
        umap_data = pd.read_csv(umap_file)
        pca_data = pd.read_csv(pca_file)
        cluster_data = pd.read_csv(cluster_file)

        # merge dimred data with clustering data
        data = ft.reduce(lambda left, right: pd.merge(left, right, on='Barcode'), [umap_data, pca_data.iloc[:, :3], cluster_data])
        data["Cluster"] = data["Cluster"].astype('category')

        # plot
        nrows = 1
        ncols = 2
        fig, axs = plt.subplots(nrows, ncols, figsize=(8*ncols, 6*nrows))
        sns.scatterplot(data=data, x="PC-1", y="PC-2", hue="Cluster", palette="tab20", ax=axs[0])
        sns.scatterplot(data=data, x="UMAP-1", y="UMAP-2", hue="Cluster", palette="tab20", ax=axs[1])
        if save is not None:
            plt.savefig(save)
        plt.show()

    def load_all(self,
                 skip: Optional[str] = None,
                 verbose: bool = False
                 ):
        # # extract read functions
        # read_funcs = [elem for elem in dir(self) if elem.startswith("load_")]
        # read_funcs = [elem for elem in read_funcs if elem not in ["load_all", "load_quicksave"]]

        for f in LOAD_FUNCS:
            if skip is None or skip not in f:
                func = getattr(self, f)
                try:
                    func(verbose=verbose)
                except ModalityNotFoundError as err:
                    if verbose:
                        print(err)

    def load_annotations(self, verbose: bool = False):
        if verbose:
            print("Loading annotations...", flush=True)
        try:
            p = self._metadata["data"]["annotations"]
        except KeyError:
            if verbose:
                raise ModalityNotFoundError(modality="annotations")
        else:
            self._annotations = read_shapesdata(path=self._path / p, mode="annotations")


    def import_annotations(self,
                           files: Optional[Union[str, os.PathLike, Path]],
                           keys: Optional[str],
                           scale_factor: Number, # µm/pixel - can be used to convert the pixel coordinates into µm coordinates
                           verbose: bool = False
                           ):
        if verbose:
            print("Importing annotations...", flush=True)

        # add annotations object
        files = convert_to_list(files)
        keys = convert_to_list(keys)

        if self._annotations is None:
            self._annotations = AnnotationsData()

        for key, file in zip(keys, files):
            # read annotation and store in dictionary
            self._annotations.add_data(data=file,
                                      key=key,
                                      scale_factor=scale_factor
                                      )

        #self._remove_empty_modalities()

    def load_regions(self, verbose: bool = False):
        if verbose:
            print("Loading regions...", flush=True)
        try:
            p = self._metadata["data"]["regions"]
        except KeyError:
            if verbose:
                raise ModalityNotFoundError(modality="regions")
        else:
            self._regions = read_shapesdata(path=self._path / p, mode="regions")

    def import_regions(self,
                    files: Optional[Union[str, os.PathLike, Path]],
                    keys: Optional[str],
                    scale_factor: Number, # µm/pixel - used to convert the pixel coordinates into µm coordinates
                    verbose: bool = False
                    ):
        if verbose:
            print("Importing regions...", flush=True)

        # add regions object
        files = convert_to_list(files)
        keys = convert_to_list(keys)
        #pixel_size = self.metadata["method_params"]['pixel_size']

        if self._regions is None:
            self._regions = RegionsData()

        for key, file in zip(keys, files):
            # read annotation and store in dictionary
            self._regions.add_data(data=file,
                                key=key,
                                scale_factor=scale_factor
                                )

        #self._remove_empty_modalities()


    def load_cells(self, verbose: bool = False):
        if verbose:
            print("Loading cells...", flush=True)

        if self._from_insitudata:
            try:
                cells_path = self._metadata["data"]["cells"]
            except KeyError:
                if verbose:
                    raise ModalityNotFoundError(modality="cells")
            else:
                self._cells = read_celldata(path=self._path / cells_path)

            # check if alt data is there and read if yes
            try:
                alt_path_dict = self._metadata["data"]["alt"]
            except KeyError:
                if verbose:
                    print("\tNo alternative cells found...")
            else:
                print("\tFound alternative cells...")
                alt_dict = {}
                for k, p in alt_path_dict.items():
                    alt_dict[k] = read_celldata(path=self._path / p)

                # add attribute
                self._alt = alt_dict
        else:
            NoProjectLoadWarning()

    def load_images(self,
                    names: Union[Literal["all", "nuclei"], str] = "all", # here a specific image can be chosen
                    overwrite: bool = False,
                    verbose: bool = False
                    ):
        # load image into ImageData object
        if verbose:
            print("Loading images...", flush=True)

        if self._from_insitudata:
            # check if matrix data is stored in this InSituData
            try:
                images_dict = self._metadata["data"]["images"]
            except KeyError:
                if verbose:
                    raise ModalityNotFoundError(modality="images")
            else:
                if names == "all":
                    img_names = list(images_dict.keys())
                else:
                    img_names = convert_to_list(names)

                # get file paths and names
                img_files = [v for k,v in images_dict.items() if k in img_names]
                img_names = [k for k,v in images_dict.items() if k in img_names]

                # create imageData object
                img_paths = [self._path / elem for elem in img_files]

                if self._images is None:
                    self._images = ImageData(img_paths, img_names)
                else:
                    for im, n in zip(img_paths, img_names):
                        self._images.add_image(im, n, overwrite=overwrite, verbose=verbose)

        else:
            NoProjectLoadWarning()

    def load_transcripts(self,
                        verbose: bool = False,
                        mode: Literal["pandas", "dask"] = "dask",
                        ):
        # read transcripts
        if verbose:
            print("Loading transcripts...", flush=True)

        if self._from_insitudata:
            # check if transcript data is stored in this InSituData
            try:
                transcripts_path = self._metadata["data"]["transcripts"]
            except KeyError:
                if verbose:
                    raise ModalityNotFoundError(modality="transcripts")
            else:
                if mode == "pandas":
                    self._transcripts = pd.read_parquet(self._path / transcripts_path)
                elif mode == "dask":
                    # Load the transcript data using Dask
                    self._transcripts = dd.read_parquet(self._path / transcripts_path)
                else:
                    raise ValueError(f"Invalid value for `mode`: {mode}")


        else:
            NoProjectLoadWarning()

    @classmethod
    def read(cls, path: Union[str, os.PathLike, Path]):
        """Read an InSituData object from a specified folder.

        Args:
            path (Union[str, os.PathLike, Path]): The path to the folder where data is saved.

        Returns:
            InSituData: A new InSituData object with the loaded data.
        """
        path = Path(path) # make sure the path is a pathlib path
        assert (path / ISPY_METADATA_FILE).exists(), "No insitupy metadata file found."
        # read InSituData metadata
        insitupy_metadata_file = path / ISPY_METADATA_FILE
        metadata = read_json(insitupy_metadata_file)

        # retrieve slide_id and sample_id
        slide_id = metadata["slide_id"]
        sample_id = metadata["sample_id"]

        # save paths of this project in metadata
        metadata["path"] = abspath(path).replace("\\", "/")
        metadata["metadata_file"] = ISPY_METADATA_FILE

        data = cls(path=path,
                   metadata=metadata,
                   slide_id=slide_id,
                   sample_id=sample_id,
                   from_insitudata=True
                   )
        return data

    def reduce_dimensions(self,
                        umap: bool = True,
                        tsne: bool = True,
                        layer: Optional[str] = None,
                        batch_correction_key: Optional[str] = None,
                        perform_clustering: bool = True,
                        verbose: bool = True,
                        tsne_lr: int = 1000,
                        tsne_jobs: int = 8,
                        **kwargs
                        ):
        """
        Reduce the dimensionality of the data using PCA, UMAP, and t-SNE techniques, optionally performing batch correction.

        Args:
            umap (bool, optional):
                If True, perform UMAP dimensionality reduction. Default is True.
            tsne (bool, optional):
                If True, perform t-SNE dimensionality reduction. Default is True.
            layer (str, optional):
                Specifies the layer of the AnnData object to operate on. Default is None (uses adata.X).
            batch_correction_key (str, optional):
                Batch key for performing batch correction using scanorama. Default is None, indicating no batch correction.
            verbose (bool, optional):
                If True, print progress messages during dimensionality reduction. Default is True.
            tsne_lr (int, optional):
                Learning rate for t-SNE. Default is 1000.
            tsne_jobs (int, optional):
                Number of CPU cores to use for t-SNE computation. Default is 8.
            **kwargs:
                Additional keyword arguments to be passed to scanorama function if batch correction is performed.

        Raises:
            ValueError: If an invalid `batch_correction_key` is provided.

        Returns:
            None: This method modifies the input matrix in place, reducing its dimensionality using specified techniques and
                batch correction if applicable. It does not return any value.
        """
        if self._cells is None:
            raise ModalityNotFoundError(modality="cells")
        else:
            cells = self._cells

        reduce_dimensions_anndata(adata=cells.matrix,
                                  umap=umap, tsne=tsne, layer=layer,
                                  batch_correction_key=batch_correction_key,
                                  perform_clustering=perform_clustering,
                                  verbose=verbose,
                                  tsne_lr=tsne_lr, tsne_jobs=tsne_jobs
                                  )

        if self._alt is not None:
            alt = self._alt
            print("Found `.alt` modality.")
            for k, cells in alt.items():
                print(f"\tReducing dimensions in `.alt['{k}']...")

                reduce_dimensions_anndata(adata=cells.matrix,
                                        umap=umap, tsne=tsne, layer=layer,
                                        batch_correction_key=batch_correction_key,
                                        perform_clustering=perform_clustering,
                                        verbose=verbose,
                                        tsne_lr=tsne_lr, tsne_jobs=tsne_jobs
                                        )

    def saveas(self,
            path: Union[str, os.PathLike, Path],
            overwrite: bool = False,
            zip_output: bool = False,
            images_as_zarr: bool = True,
            zarr_zipped: bool = False,
            images_max_resolution: Optional[Number] = None, # in µm per pixel
            verbose: bool = True
            ):
        '''
        Function to save the InSituData object.

        Args:
            path: Path to save the data to.
        '''
        # check if the path already exists
        path = Path(path)

        # check overwrite
        check_overwrite_and_remove_if_true(path=path, overwrite=overwrite)

        if zip_output:
            zippath = path / (path.stem + ".zip")
            check_overwrite_and_remove_if_true(path=zippath, overwrite=overwrite)

        print(f"Saving data to {str(path)}") if verbose else None

        # create output directory if it does not exist yet
        path.mkdir(parents=True, exist_ok=True)

        # store basic information about experiment
        self._metadata["slide_id"] = self._slide_id
        self._metadata["sample_id"] = self._sample_id

        # clean old entries in data metadata
        self._metadata["data"] = {}

        # save images
        if self._images is not None:
            images = self._images
            _save_images(
                imagedata=images,
                path=path,
                metadata=self._metadata,
                images_as_zarr=images_as_zarr,
                zipped=zarr_zipped,
                max_resolution=images_max_resolution,
                verbose=False
                )

        # save cells
        if self._cells is not None:
            cells = self._cells
            _save_cells(
                cells=cells,
                path=path,
                metadata=self._metadata,
                boundaries_zipped=zarr_zipped
            )

        # save alternative cell data
        if self._alt is not None:
            alt = self._alt
            _save_alt(
                attr=alt,
                path=path,
                metadata=self._metadata,
                boundaries_zipped=zarr_zipped
            )

        # save transcripts
        if self._transcripts is not None:
            transcripts = self._transcripts
            _save_transcripts(
                transcripts=transcripts,
                path=path,
                metadata=self._metadata
                )

        # save annotations
        if self._annotations is not None:
            annotations = self._annotations
            _save_annotations(
                annotations=annotations,
                path=path,
                metadata=self._metadata
            )

        # save regions
        if self._regions is not None:
            regions = self._regions
            _save_regions(
                regions=regions,
                path=path,
                metadata=self._metadata
            )

        # save version of InSituPy
        self._metadata["version"] = __version__

        if "method_params" in self._metadata:
            # move method_param key to end of metadata
            self._metadata["method_params"] = self._metadata.pop("method_params")

        # write Xeniumdata metadata to json file
        xd_metadata_path = path / ISPY_METADATA_FILE
        write_dict_to_json(dictionary=self._metadata, file=xd_metadata_path)

        # Optionally: zip the resulting directory
        if zip_output:
            shutil.make_archive(path, 'zip', path, verbose=False)
            shutil.rmtree(path) # delete directory

        # change path to the new one
        self._path = path.resolve()

        # reload the modalities
        self.reload(verbose=False)

        print("Saved.") if verbose else None

    def save(self,
             path: Optional[Union[str, os.PathLike, Path]] = None,
             zarr_zipped: bool = False,
             verbose: bool = True,
             keep_history: bool = False
             ):

        # check path
        if path is not None:
            path = Path(path)
        else:
            if self._from_insitudata:
                path = Path(self._metadata["path"])
            else:
                warn(
                    f"Data as not loaded from an InSituPy project. "
                    f"Use `saveas()` instead to save the data to a new project folder."
                    )

        if path.exists():
            # check if path is a valid directory
            if not path.is_dir():
                raise NotADirectoryError(f"Path is not a directory: {str(path)}")

            # check if the folder is a InSituPy project
            metadata_file = path / ISPY_METADATA_FILE

            if metadata_file.exists():
                # read metadata file and check uid
                project_meta = read_json(metadata_file)

                # check uid
                project_uid = project_meta["uids"][-1]  # [-1] to select latest uid
                current_uid = self._metadata["uids"][-1]
                if current_uid == project_uid:
                    self._update_to_existing_project(path=path,
                                                     zarr_zipped=zarr_zipped,
                                                     verbose=verbose
                                                     )

                    # reload the modalities
                    self.reload(verbose=False, skip=["transcripts", "images"])

                    if not keep_history:
                        self.remove_history(verbose=False)
                else:
                    warn(
                        f"UID of current object {current_uid} not identical with UID in project path {path}: {project_uid}.\n"
                        f"Project is neither saved nor updated. Try `saveas()` instead to save the data to a new project folder. "
                        f"A reason for this could be the data has been cropped in the meantime."
                    )
            else:
                warn(
                    f"No `.ispy` metadata file in {path}. Directory is probably no valid InSituPy project. "
                    f"Use `saveas()` instead to save the data to a new InSituPy project."
                    )


        else:
            # save to the respective directory
            self.saveas(path=path)

    def save_current_colorlegend(self, savepath):

        # Check if static_canvas exists
        if not hasattr(config, 'static_canvas'):
            print("Warning: 'static_canvas' attribute not found in config. "
                "Please display data in the napari viewer using '.show()' first.")
            return

        try:
            # Save the figure to a PDF file
            config.static_canvas.figure.savefig(savepath)
            print(f"Figure saved as {savepath}")
        except RuntimeError as e:
            if 'FigureCanvasQTAgg has been deleted' in str(e):
                print("Warning: The color legend has been deleted and cannot be saved.")
            else:
                raise  # Re-raise the exception if it's a different error

    def _update_to_existing_project(self,
                                    path: Optional[Union[str, os.PathLike, Path]],
                                    zarr_zipped: bool = False,
                                    verbose: bool = True
                                    ):
        if verbose:
            print(f"Updating project in {path}")

        # save cells
        if self._cells is not None:
            cells = self._cells
            if verbose:
                print("\tUpdating cells...", flush=True)
            _save_cells(
                cells=cells,
                path=path,
                metadata=self._metadata,
                boundaries_zipped=zarr_zipped,
                overwrite=True
            )

        # save alternative cell data
        if self._alt is not None:
            alt = self._alt
            if verbose:
                print("\tUpdating alternative segmentations...", flush=True)
            _save_alt(
                attr=alt,
                path=path,
                metadata=self._metadata,
                boundaries_zipped=zarr_zipped
            )

        # save annotations
        if self._annotations is not None:
            annotations = self._annotations
            if verbose:
                print("\tUpdating annotations...", flush=True)
            _save_annotations(
                annotations=annotations,
                path=path,
                metadata=self._metadata
            )

        # save regions
        if self._regions is not None:
            regions = self._regions
            if verbose:
                print("\tUpdating regions...", flush=True)
            _save_regions(
                regions=regions,
                path=path,
                metadata=self._metadata
            )

        # save version of InSituPy
        self._metadata["version"] = __version__

        if "method_params" in self._metadata:
            # move method_params key to end of metadata
            self._metadata["method_params"] = self._metadata.pop("method_params")

        # write Xeniumdata metadata to json file
        xd_metadata_path = path / ISPY_METADATA_FILE
        write_dict_to_json(dictionary=self._metadata, file=xd_metadata_path)

        if verbose:
            print("Saved.")


    def quicksave(self,
                  note: Optional[str] = None
                  ):
        # create quicksave directory if it does not exist already
        self._quicksave_dir = CACHE / "quicksaves"
        self._quicksave_dir.mkdir(parents=True, exist_ok=True)

        # save annotations
        if self._annotations is None:
            print("No annotations found. Quicksave skipped.", flush=True)
        else:
            annotations = self._annotations
            # create filename
            current_datetime = datetime.now().strftime("%y%m%d_%H-%M-%S")
            slide_id = self._slide_id
            sample_id = self._sample_id
            uid = str(uuid4())[:8]

            # create output directory
            outname = f"{slide_id}__{sample_id}__{current_datetime}__{uid}"
            outdir = self._quicksave_dir / outname

            _save_annotations(
                annotations=annotations,
                path=outdir,
                metadata=None
            )

            if note is not None:
                with open(outdir / "note.txt", "w") as notefile:
                    notefile.write(note)

            # # # zip the output
            # shutil.make_archive(outdir, format='zip', root_dir=outdir, verbose=False)
            # shutil.rmtree(outdir) # delete directory


    def list_quicksaves(self):
        pattern = "{slide_id}__{sample_id}__{savetime}__{uid}"

        # collect results
        res = {
            "slide_id": [],
            "sample_id": [],
            "savetime": [],
            "uid": [],
            "note": []
        }
        for d in self._quicksave_dir.glob("*"):
            parse_res = parse(pattern, d.stem).named
            for key, value in parse_res.items():
                res[key].append(value)

            notepath = d / "note.txt"
            if notepath.exists():
                with open(notepath, "r") as notefile:
                    res["note"].append(notefile.read())
            else:
                res["note"].append("")

        # create and return dataframe
        return pd.DataFrame(res)

    def load_quicksave(self,
                       uid: str
                       ):
        # find files with the uid
        files = list(self._quicksave_dir.glob(f"*{uid}*"))

        if len(files) == 1:
            ad = read_shapesdata(files[0] / "annotations", mode="annotations")
        elif len(files) == 0:
            print(f"No quicksave with uid '{uid}' found. Use `.list_quicksaves()` to list all available quicksaves.")
        else:
            raise ValueError(f"More than one quicksave with uid '{uid}' found.")

        # add annotations to existing annotations attribute or add a new one
        if self._annotations is None:
            self._annotations = AnnotationsData()
        else:
            annotations = self._annotations
            for k in ad.metadata.keys():
                annotations.add_data(ad[k], k, verbose=True)


    def show(self,
        keys: Optional[str] = None,
        # annotation_keys: Optional[str] = None,
        point_size: int = 8,
        scalebar: bool = True,
        #pixel_size: float = None, # if none, extract from metadata
        unit: str = "µm",
        # cmap_annotations: str ="Dark2",
        grayscale_colormap: List[str] = ["red", "green", "cyan", "magenta", "yellow", "gray"],
        return_viewer: bool = False,
        widgets_max_width: int = 500
        ):

        # create viewer
        self._viewer = napari.Viewer(title=f"{self._slide_id}: {self._sample_id}")

        if self._images is None:
            warn("No attribute `.images` found.")
        else:
            images_attr = self._images
            n_images = len(images_attr.metadata)
            n_grayscales = 0 # number of grayscale images
            for i, (img_name, img_metadata) in enumerate(images_attr.metadata.items()):
            #for i, img_name in enumerate(image_keys):
                img = images_attr[img_name]
                is_visible = False if i < n_images - 1 else True # only last image is set visible
                pixel_size = img_metadata['pixel_size']

                # check if the current image is RGB
                is_rgb = self._images.metadata[img_name]["rgb"]

                if is_rgb:
                    cmap = None  # default value of cmap
                    blending = "translucent_no_depth"  # set blending mode
                else:
                    if img_name == "nuclei":
                        cmap = "blue"
                    else:
                        cmap = grayscale_colormap[n_grayscales]
                        n_grayscales += 1
                    blending = "additive"  # set blending mode


                if not isinstance(img, list):
                    # create image pyramid for lazy loading
                    img_pyramid = create_img_pyramid(img=img, nsubres=6)
                else:
                    img_pyramid = img

                # infer contrast limits
                contrast_limits = _get_contrast_limits(img_pyramid)

                if contrast_limits[1] == 0:
                    warn("The maximum value of the image is 0. Is the image really completely empty?")
                    contrast_limits = (0, 255)

                # add img pyramid to napari viewer
                self._viewer.add_image(
                        img_pyramid,
                        name=img_name,
                        colormap=cmap,
                        blending=blending,
                        rgb=is_rgb,
                        contrast_limits=contrast_limits,
                        scale=(pixel_size, pixel_size),
                        visible=is_visible
                    )

        # optionally: add cells as points
        #if show_cells or keys is not None:
        if keys is not None:
            if self._cells is None:
                raise InSituDataMissingObject("cells")
            else:
                cells = self._cells
                # convert keys to list
                keys = convert_to_list(keys)

                # get point coordinates
                points = np.flip(cells.matrix.obsm["spatial"].copy(), axis=1) # switch x and y (napari uses [row,column])
                #points *= pixel_size # convert to length unit (e.g. µm)

                # get expression matrix
                if issparse(cells.matrix.X):
                    X = cells.matrix.X.toarray()
                else:
                    X = cells.matrix.X

                for i, k in enumerate(keys):
                    #pvis = False if i < len(keys) - 1 else True # only last image is set visible
                    # get expression values
                    if k in cells.matrix.obs.columns:
                        color_value = cells.matrix.obs[k].values

                    else:
                        geneid = cells.matrix.var_names.get_loc(k)
                        color_value = X[:, geneid]

                    # extract names of cells
                    cell_names = cells.matrix.obs_names.values

                    # create points layer
                    layer = _create_points_layer(
                        points=points,
                        color_values=color_value,
                        name=k,
                        point_names=cell_names,
                        point_size=point_size,
                        visible=True
                    )

                    # add layer programmatically - does not work for all types of layers
                    # see: https://forum.image.sc/t/add-layerdatatuple-to-napari-viewer-programmatically/69878
                    self._viewer.add_layer(Layer.create(*layer))

        # WIDGETS
        if self._cells is None:
            # add annotation widget to napari
            add_geom_widget = add_new_geometries_widget()
            add_geom_widget.max_height = 120
            add_geom_widget.max_width = widgets_max_width
            self._viewer.window.add_dock_widget(add_geom_widget, name="Add geometries", area="right")
        else:
            cells = self._cells
            # initialize the widgets
            show_points_widget, locate_cells_widget, show_geometries_widget, show_boundaries_widget, select_data, filter_cells_widget = _initialize_widgets(xdata=self)

            # add widgets to napari window
            if select_data is not None:
                self._viewer.window.add_dock_widget(select_data, name="Select data", area="right")
                select_data.max_height = 50
                select_data.max_width = widgets_max_width

            if show_points_widget is not None:
                self.viewer.window.add_dock_widget(show_points_widget, name="Show data", area="right")
                show_points_widget.max_height = 170
                show_points_widget.max_width = widgets_max_width



            if show_boundaries_widget is not None:
                self._viewer.window.add_dock_widget(show_boundaries_widget, name="Show boundaries", area="right")
                #show_boundaries_widget.max_height = 80
                show_boundaries_widget.max_width = widgets_max_width

            if locate_cells_widget is not None:
                self._viewer.window.add_dock_widget(locate_cells_widget, name="Navigate to cell", area="right", tabify=True)
                #locate_cells_widget.max_height = 130
                locate_cells_widget.max_width = widgets_max_width

            if filter_cells_widget is not None:
                self.viewer.window.add_dock_widget(filter_cells_widget, name="Filter cells", area="right", tabify=True)
                filter_cells_widget.max_height = 150
                show_points_widget.max_width = widgets_max_width

            # add annotation widget to napari
            add_geom_widget = add_new_geometries_widget()
            #annot_widget.max_height = 100
            add_geom_widget.max_width = widgets_max_width
            self._viewer.window.add_dock_widget(add_geom_widget, name="Add geometries", area="right", add_vertical_stretch=True)

            # if show_region_widget is not None:
            #     self.viewer.window.add_dock_widget(show_region_widget, name="Show regions", area="right")
            #     show_region_widget.max_height = 100
            #     show_region_widget.max_width = widgets_max_width

            if show_geometries_widget is not None:
                self._viewer.window.add_dock_widget(show_geometries_widget, name="Show geometries", area="right", tabify=True)
                show_geometries_widget.max_width = widgets_max_width

        # EVENTS
        # Assign function to an layer addition event
        def _update_uid(event):
            if event is not None:

                layer = event.source
                if event.action == "add":
                    if 'uid' in layer.properties:
                        layer.properties['uid'][-1] = str(uuid4())
                    else:
                        layer.properties['uid'] = np.array([str(uuid4())], dtype='object')

                elif event.action == "remove":
                    pass
                else:
                    raise ValueError("Unexpected value '{event.action}' for `event.action`. Expected 'add' or 'remove'.")

        # Assign the function to data of all existing layers
        for layer in self._viewer.layers:
            if isinstance(layer, Shapes) or isinstance(layer, Points):
                layer.events.data.connect(_update_uid)

        # Connect the function to the data of existing shapes and points layers in the viewer
        def connect_to_all_shapes_layers(event):
            layer = event.source[event.index]
            if event is not None:
                if isinstance(layer, Shapes) or isinstance(layer, Points):
                    layer.events.data.connect(_update_uid)

        # Connect the function to any new layers added to the viewer
        self._viewer.layers.events.inserted.connect(connect_to_all_shapes_layers)

        # add color legend widget
        import insitupy._core.config as config
        from insitupy._core.config import init_colorlegend_canvas
        init_colorlegend_canvas()
        self._viewer.window.add_dock_widget(config.static_canvas, area='left', name='Color legend')

        # def update_colorlegend(event):
        #     # if event.type == "inserted":
        #     layer = event.source[event.index]
        #     _add_colorlegend_to_canvas(layer=layer, static_canvas=config.static_canvas)
        #     # if event.type == "removed":
        #         # config.static_canvas.figure.clear()
        #         # config.static_canvas.draw()

        # self.viewer.layers.events.inserted.connect(update_colorlegend)
        # #self.viewer.layers.events.removed.connect(add_colorlegend_widget)

        # NAPARI SETTINGS
        if scalebar:
            # add scale bar
            self._viewer.scale_bar.visible = True
            self._viewer.scale_bar.unit = unit

        napari.run()
        if return_viewer:
            return self._viewer

    def store_geometries(self,
                         name_pattern = "{type_symbol} {class_name} ({annot_key})",
                         uid_col: str = "id"
                         ):
        """
        Extracts geometric layers from shapes and points layers in the napari viewer
        and stores them in the InSituData object as annotations or regions.

        Args:
            name_pattern (str): A format string used to parse the layer names.
                It should contain placeholders for 'type_symbol', 'class_name',
                and 'annot_key'.
            uid_col (str): The name of the column used to store unique identifiers
                for the geometries. Default is "id".

        Raises:
            AttributeError: If the viewer is not initialized, an error message
                prompts the user to open a napari viewer using the `.show()` method.

        Notes:
            - The function iterates through the layers in the viewer and checks if
            they are instances of Shapes or Points.
            - It extracts the geometric data, colors, and other relevant properties
            to create a GeoDataFrame.
            - The GeoDataFrame is then added to the annotations or regions of the
            InSituData object based on the type of layer.
            - If the layer is classified as a region but is a point layer, a warning
            is issued, and the layer is skipped.
        """
        if self._viewer is not None:
            viewer = self._viewer
        else:
            print("Use `.show()` first to open a napari viewer.")

        # iterate through layers and save them as annotation or region if they meet requirements
        layers = viewer.layers
        #collection_dict = {}
        for layer in layers:
            if isinstance(layer, Shapes) or isinstance(layer, Points):
                name_parsed = parse(name_pattern, layer.name)
                if name_parsed is not None:
                    type_symbol = name_parsed.named["type_symbol"]
                    annot_key = name_parsed.named["annot_key"]
                    class_name = name_parsed.named["class_name"]

                    # if the InSituData object does not have an annotations attribute, initialize it
                    if self._annotations is None:
                        self._annotations = AnnotationsData() # initialize empty object

                    # extract shapes coordinates and colors
                    layer_data = layer.data
                    colors = layer.edge_color.tolist()
                    scale = layer.scale

                    checks_passed = True
                    is_region_layer = False
                    object_type = "annotation"
                    if type_symbol == REGIONS_SYMBOL:
                        is_region_layer = True
                        object_type = "region"
                        if isinstance(layer, Points):
                            warn(f'Layer "{layer.name}" is a point layer and at the same time classified as "Region". This is not allowed. Skipped this layer.')
                            checks_passed = False

                    if object_type == "annotation":
                        # if the InSituData object does not have an annotations attribute, initialize it
                        if not hasattr(self, "annotations"):
                            self.annotations = AnnotationsData() # initialize empty object
                    else:
                        # if the InSituData object does not have an regions attribute, initialize it
                        if not hasattr(self, "regions"):
                            self.regions = RegionsData() # initialize empty object

                    if checks_passed:
                        if isinstance(layer, Shapes):
                            # extract shape types
                            shape_types = layer.shape_type
                            # build annotation GeoDataFrame
                            geom_df = {
                                uid_col: layer.properties["uid"],
                                "objectType": object_type,
                                #"geometry": [Polygon(np.stack([ar[:, 1], ar[:, 0]], axis=1)) for ar in layer_data],  # switch x/y
                                "geometry": [convert_napari_shape_to_polygon_or_line(napari_shape_data=ar, shape_type=st) for ar, st in zip(layer_data, shape_types)],
                                "name": class_name,
                                "color": [[int(elem[e]*255) for e in range(3)] for elem in colors],
                                #"scale": [scale] * len(layer_data),
                                #"layer_type": ["Shapes"] * len(layer_data)
                            }

                        elif isinstance(layer, Points):
                            # build annotation GeoDataFrame
                            geom_df = {
                                uid_col: layer.properties["uid"],
                                "objectType": object_type,
                                "geometry": [Point(d[1], d[0]) for d in layer_data],  # switch x/y
                                "name": class_name,
                                "color": [[int(elem[e]*255) for e in range(3)] for elem in colors],
                                #"scale": [scale] * len(layer_data),
                                #"layer_type": ["Points"] * len(layer_data)
                            }

                        # generate GeoDataFrame
                        geom_df = GeoDataFrame(geom_df, geometry="geometry")

                        if is_region_layer:
                            if self._regions is None:
                                self._regions = RegionsData()

                            # add regions
                            self._regions.add_data(data=geom_df,
                                                  key=annot_key,
                                                  verbose=True,
                                                  scale_factor=scale[0]
                                                  )
                        else:
                            if self._annotations is None:
                                self._annotations = AnnotationsData()

                            # add annotations
                            self._annotations.add_data(data=geom_df,
                                                      key=annot_key,
                                                      verbose=True,
                                                      scale_factor=scale[0]
                                                      )

            else:
                pass

        #self._remove_empty_modalities()

    def plot_binned_expression(
        self,
        genes: Union[List[str], str],
        maxcols: int = 4,
        figsize: Tuple[int, int] = (8,6),
        savepath: Union[str, os.PathLike, Path] = None,
        save_only: bool = False,
        dpi_save: int = 300,
        show: bool = True,
        fontsize: int = 28
        ):
        # extract binned expression matrix and gene names
        binex = self._cells.matrix.varm["binned_expression"]
        gene_names = self._cells.matrix.var_names

        genes = convert_to_list(genes)

        nplots, nrows, ncols = get_nrows_maxcols(len(genes), max_cols=maxcols)

        # setup figure
        fig, axs = plt.subplots(nrows, ncols, figsize=(figsize[0]*ncols, figsize[1]*nrows))

        # scale font sizes
        plt.rcParams.update({'font.size': fontsize})

        if nplots > 1:
            axs = axs.ravel()
        else:
            axs = [axs]

        for i, gene in enumerate(genes):
            # retrieve binned expression
            img = binex[gene_names.get_loc(gene)]

            # determine upper limit for color
            vmax = np.percentile(img[img>0], 95)

            # plot expression
            axs[i].imshow(img, cmap="viridis", vmax=vmax)

            # set title
            axs[i].set_title(gene)

        if nplots > 1:

            # check if there are empty plots remaining
            while i < nrows * maxcols - 1:
                i+=1
                # remove empty plots
                axs[i].set_axis_off()

        if show:
            fig.tight_layout()
            save_and_show_figure(savepath=savepath, fig=fig, save_only=save_only, dpi_save=dpi_save)
        else:
            return fig, axs

    def plot_expr_along_obs_val(
        self,
        keys: str,
        obs_val: str,
        groupby: Optional[str] = None,
        method: Literal["lowess", "loess"] = 'loess',
        stderr: bool = False,
        savepath=None,
        return_data=False,
        **kwargs
        ):
        # retrieve anndata object from InSituData
        adata = self._cells.matrix

        results = expr_along_obs_val(
            adata=adata,
            keys=keys,
            obs_val=obs_val,
            groupby=groupby,
            method=method,
            stderr=stderr,
            savepath=savepath,
            return_data=return_data
            **kwargs
            )

        if return_data:
            return results

    def reload(
        self,
        skip: Optional[List] = None,
        verbose: bool = True
        ):
        data_meta = self._metadata["data"]
        current_modalities = [m for m in MODALITIES if getattr(self, m) is not None and m in data_meta]

        if skip is not None:
            # remove the modalities which are supposed to be skipped during reload
            skip = convert_to_list(skip)
            for s in skip:
                try:
                    current_modalities.remove(s)
                except ValueError:
                    pass

        if len(current_modalities) > 0:
            print(f"Reloading following modalities: {', '.join(current_modalities)}") if verbose else None
            for cm in current_modalities:
                func = getattr(self, f"load_{cm}")
                func(verbose=verbose)
        else:
            print("No modalities with existing save path found. Consider saving the data with `saveas()` first.")

    def remove_history(self,
                       verbose: bool = True
                       ):

        for cat in ["annotations", "cells", "regions"]:
            dirs_to_remove = []
            #if hasattr(self, cat):
            files = sorted((self._path / cat).glob("*"))
            if len(files) > 1:
                dirs_to_remove = files[:-1]

                for d in dirs_to_remove:
                    shutil.rmtree(d)

                print(f"Removed {len(dirs_to_remove)} entries from '.{cat}'.") if verbose else None
            else:
                print(f"No history found for '{cat}'.") if verbose else None

    def remove_modality(self,
                        modality: str
                        ):
        if hasattr(self, modality):
            # delete attribute from InSituData object
            delattr(self, modality)

            # delete metadata
            self.metadata["data"].pop(modality, None) # returns None if key does not exist

        else:
            print(f"No modality '{modality}' found. Nothing removed.")


def calc_distance_of_cells_from(
    data: InSituData,
    annotation_key: str,
    annotation_class: str,
    region_key: Optional[str] = None,
    region_name: Optional[str] = None,
    key_to_save: Optional[str] = None
    ):

    """
    Calculate the distance of cells from a specified annotation class within a given region and save the results.

    This function calculates the distance of each cell in the spatial data to the closest point
    of a specified annotation class. The distances are then saved in the cell data matrix.

    Args:
        data (InSituData): The input data containing cell and annotation information.
        annotation_key (str): The key to retrieve the annotation information.
        annotation_class (Optional[str]): The specific annotation class to calculate distances from.
        region_key: (Optional[str]): If not None, `region_key` is used together with `region_name` to determine the region in which cells are considered
                                     for the analysis.
        region_name: (Optional[str]): If not None, `region_name` is used together with `region_key` to determine the region in which cells are considered
                                     for the analysis.
        key_to_save (Optional[str]): The key under which to save the calculated distances in the cell data matrix.
                                     If None, a default key is generated based on the annotation class.

    Returns:
        None
    """
    # extract anndata object
    adata = data.cells.matrix

    if region_name is None:
        print(f'Calculate the distance of cells from the annotation "{annotation_class}"')
        region_mask = [True] * len(adata)
    else:
        assert region_key is not None, "`region_key` must not be None if `region_name` is not None."
        print(f'Calculate the distance of cells from the annotation "{annotation_class}" within region "{region_name}"')

        try:
            region_df = adata.obsm["regions"]
        except KeyError:
            data.assign_regions(keys=region_key)
            region_df = adata.obsm["regions"]
        else:
            if region_key not in region_df.columns:
                data.assign_regions(keys=region_key)

        # generate mask for selected region
        region_mask = region_df[region_key] == region_name

    # create geopandas points from cells
    x = adata.obsm["spatial"][:, 0][region_mask]
    y = adata.obsm["spatial"][:, 1][region_mask]
    indices = adata.obs_names[region_mask]
    cells = gpd.points_from_xy(x, y)

    # retrieve annotation information
    annot_df = data.annotations[annotation_key]
    class_df = annot_df[annot_df["name"] == annotation_class]

    # calculate distance of cells to their closest point
    # scaled_geometries = [
    #     scale_func(geometry, xfact=scale[0], yfact=scale[1], origin=(0,0))
    #     for geometry, scale in zip(class_df["geometry"], class_df["scale"])
    #     ]
    scaled_geometries = class_df["geometry"].tolist()
    dists = np.array([cells.distance(geometry) for geometry in scaled_geometries])
    min_dists = dists.min(axis=0)

    # add indices to minimum distances
    min_dists = pd.Series(min_dists, index=indices)

    # add results to CellData
    if key_to_save is None:
        #key_to_save = f"dist_from_{annotation_class}"
        key_to_save = annotation_class
    #adata.obs[key_to_save] = min_dists

    obsm_keys = adata.obsm.keys()
    if "distance_from" not in obsm_keys:
        # add empty pandas dataframe with obs_names as index
        adata.obsm["distance_from"] = pd.DataFrame(index=adata.obs_names)

    adata.obsm["distance_from"][key_to_save] = min_dists
    print(f'Saved distances to `.cells.matrix.obsm["distance_from"]["{key_to_save}"]`')

from insitupy.utils._dge import _select_data_for_dge, _substitution_func


def differential_gene_expression(
    target: InSituData,
    target_annotation_tuple: Optional[Tuple[str, str]] = None,
    target_cell_type_tuple: Optional[Tuple[str, str]] = None,
    target_region_tuple: Optional[Tuple[str, str]] = None,
    ref: Optional[Union[InSituData, List[InSituData]]] = None,
    ref_annotation_tuple: Optional[Union[Literal["rest", "same"], Tuple[str, str]]] = "same",
    ref_cell_type_tuple: Optional[Union[Literal["rest", "same"], Tuple[str, str]]] = "same",
    ref_region_tuple: Optional[Tuple[str, str]] = "same",
    significance_threshold: Number = 0.05,
    fold_change_threshold: Number = 1,
    plot_volcano: bool = True,
    return_results: bool = False,
    method: Optional[Literal['logreg', 't-test', 'wilcoxon', 't-test_overestim_var']] = 't-test',
    exclude_ambiguous_assignments: bool = False,
    force_assignment: bool = False,
    title: Optional[str] = None,
    savepath: Union[str, os.PathLike, Path] = None,
    save_only: bool = False,
    dpi_save: int = 300,
    verbose: bool = False,
    **volcano_kwargs
):
    """
    Perform differential gene expression analysis on in situ sequencing data.

    This function compares gene expression between specified annotations within a single
    InSituData object or between two InSituData objects. It supports various statistical
    methods for differential expression analysis and can generate a volcano plot of the results.

    Args:
        target (InSituData): The primary in situ data object.
        target_annotation_tuple (Optional[Tuple[str, str]]): Tuple containing the annotation key and name for the target data.
        target_cell_type_tuple (Optional[Tuple[str, str]]): Tuple specifying an observation key and value to filter the target data by cell type.
        target_region_tuple (Optional[Tuple[str, str]]): Tuple specifying a region key and name to restrict the analysis to a specific region in the target data.
        ref (Optional[Union[InSituData, List[InSituData]]]): Reference in situ data object(s) for comparison. Defaults to None.
        ref_annotation_tuple (Optional[Union[Literal["rest", "same"], Tuple[str, str]]]): Tuple containing the reference annotation key and name, or "rest" to use the rest of the data as reference, or "same" to use the same annotation as the target. Defaults to "same".
        ref_cell_type_tuple (Optional[Union[Literal["rest", "same"], Tuple[str, str]]]): Tuple specifying an observation key and value to filter the reference data by cell type, or "rest" to use the rest of the data, or "same" to use the same cell type as the target. Defaults to "same".
        ref_region_tuple (Optional[Tuple[str, str]]): Tuple specifying a region key and name to restrict the analysis to a specific region in the reference data. Defaults to None.
        significance_threshold (float): P-value threshold for significance (default is 0.05).
        fold_change_threshold (float): Log2 fold change threshold for up/down regulation (default is 1).
        plot_volcano (bool): Whether to generate a volcano plot of the results. Defaults to True.
        return_results (bool): Whether to return the results as dictionary including the dataframe differentially expressed genes and the parameters.
        method (Optional[Literal['logreg', 't-test', 'wilcoxon', 't-test_overestim_var']]): Statistical method to use for differential expression analysis. Defaults to 't-test'.
        exclude_ambiguous_assignments (bool): Whether to exclude ambiguous assignments in the data. Defaults to False.
        force_assignment (bool): Whether to force assignment of annotations and regions even if it has been done before already. Defaults to False.
        title (Optional[str]): Title for the volcano plot. Defaults to None.
        savepath (Union[str, os.PathLike, Path]): Path to save the plot. Defaults to None.
        save_only (bool): If True, only save the plot without displaying it. Defaults to False.
        dpi_save (int): Dots per inch (DPI) for saving the plot. Defaults to 300.
        verbose (bool): Whether to print detailed information during the analysis. Defaults to False.
        **volcano_kwargs: Additional keyword arguments for the volcano plot.

    Returns:
        Union[None, Dict[str, Any]]: If `plot_volcano` is True, returns None. Otherwise, returns a dictionary with the results DataFrame and parameters used for the analysis.

    Raises:
        ValueError: If `ref_annotation_tuple` is neither 'rest' nor a 2-tuple.
        AssertionError: If `ref` is provided when `ref_annotation_tuple` is 'rest'.
        AssertionError: If `target_region_tuple` is provided when `ref` is not None.
        AssertionError: If the specified region or annotation is not found in the data.

    Example:
        >>> result = differential_gene_expression(
                target=my_data,
                target_annotation_tuple=("pathologist", "tumor"),
                ref=my_ref_data,
                ref_annotation_tuple=("cell_type", "astrocyte"),
                plot_volcano=True,
                method='wilcoxon'
            )
    """
    if not (plot_volcano | return_results):
        raise ValueError("Both `plot_volcano` and `return_results` are False. At least one of them must be True.")

    dge_comparison_column = "DGE_COMPARISON_COLUMN"

    # pre-flight checks
    if ref_annotation_tuple is not None:
        if ref_annotation_tuple == "rest":
            if ref is not None:
                raise ValueError("Value 'rest' for `ref_annotation_tuple` is only allowed if no reference data is given (`ref=None`).")
        elif ref_annotation_tuple == "same":
            ref_annotation_tuple = target_annotation_tuple
        elif not isinstance(ref_annotation_tuple, tuple):
            raise ValueError(f"Unknown type of `ref_annotation_tuple`: {type(ref_annotation_tuple)}. Must be either tuple, 'rest', 'same' or None.")
        else:
            pass

    if ref_region_tuple is not None:
        if ref_region_tuple == "rest":
            if ref is not None:
                raise ValueError("Value 'rest' for `ref_region_tuple` is only allowed if no reference data is given (`ref=None`).")
        elif ref_region_tuple == "same":
            ref_region_tuple = target_region_tuple
        elif not isinstance(ref_region_tuple, tuple):
            raise ValueError(f"Unknown type of `ref_region_tuple`: {type(ref_region_tuple)}. Must be either tuple, 'rest', 'same' or None.")
        else:
            pass

    if ref_cell_type_tuple is not None:
        if ref_cell_type_tuple == "rest":
            if ref is not None:
                raise ValueError("Value 'rest' for `ref_cell_type_tuple` is only allowed if no reference data is given (`ref=None`).")
        elif ref_cell_type_tuple == "same":
            ref_cell_type_tuple = target_cell_type_tuple
        elif not isinstance(ref_cell_type_tuple, tuple):
            raise ValueError(f"Unknown type of `ref_cell_type_tuple`: {type(ref_cell_type_tuple)}. Must be either tuple, 'rest', 'same' or None.")
        else:
            pass

    # select data for analysis
    adata_data = _select_data_for_dge(
        data=target,
        annotation_tuple=target_annotation_tuple,
        cell_type_tuple=target_cell_type_tuple,
        region_tuple=target_region_tuple,
        force_assignment=force_assignment,
        verbose=verbose
    )

    # original tuples for plotting the configuration table
    orig_ref_annotation_tuple = ref_annotation_tuple
    orig_ref_cell_type_tuple = ref_cell_type_tuple

    if ref is None:
        ref = target.copy()

        # TODO: Implement behavior for "rest"
        # The "rest" argument is only implemented if ref_data is None in the beginning
        if ref_annotation_tuple == "rest":
            rest_annotations = [
                elem
                for elem in ref.cells.matrix.obsm["annotations"][target_annotation_tuple[0]].unique()
                if elem != target_annotation_tuple[1]
                ]
            ref_annotation_tuple = (target_annotation_tuple[0], rest_annotations)

        if ref_region_tuple == "rest":
            rest_regions = [
                elem
                for elem in ref.cells.matrix.obsm["regions"][target_region_tuple[0]].unique()
                if elem != target_region_tuple[1]
                ]
            ref_region_tuple = (target_region_tuple[0], rest_regions)

        if ref_cell_type_tuple == "rest":
            rest_cell_types = [
                elem
                for elem in ref.cells.matrix.obs[target_cell_type_tuple[0]].unique()
                if elem != target_cell_type_tuple[1]
                ]
            ref_cell_type_tuple = (target_cell_type_tuple[0], rest_cell_types)

    if isinstance(ref, InSituData):
        # generate a list from ref_dta
        ref = [ref]
    elif isinstance(ref, list):
        assert np.all([isinstance(elem, InSituData) for elem in ref]), "Not all elements of list given in `ref` are InSituData objects."
    else:
        raise ValueError("`ref` must be an InSituData object or a list of InSituData objects.")

    adata_ref_list = []
    for rd in ref:
        # select reference data for analysis
        ad_ref = _select_data_for_dge(
            data=rd,
            annotation_tuple=ref_annotation_tuple,
            cell_type_tuple=ref_cell_type_tuple,
            region_tuple=ref_region_tuple,
            force_assignment=force_assignment,
            verbose=verbose
        )
        adata_ref_list.append(ad_ref)

    if len(adata_ref_list) > 1:
        adata_ref = anndata.concat(adata_ref_list)
    else:
        adata_ref = adata_ref_list[0]

    # check before concatenation whether cells with identical names are found in both data and reference
    if not set(adata_data.obs_names).isdisjoint(set(adata_ref.obs_names)):
        n_duplicated_cells = len(set(adata_data.obs_names).intersection(set(adata_ref.obs_names)))
        pct_duplicated_cells = round((n_duplicated_cells / 2) / (len(adata_data) + len(adata_data)) * 100, 1)

        warn(
            f"{n_duplicated_cells} ({pct_duplicated_cells}%) cells were found to belong to both data and reference. "
            "This can happen due to overlapping annotations or non-unique cell names in the individual datasets. "
            "If you are sure that the same cell cannot be found in both data and reference, you can ignore this warning. "
            "To exclude ambiguously assigned cells from the analysis, use `exclude_ambiguous_assignments=True`."
        )

    # concatenate and ignore user warning about observations being not unique since we take care of this later by filtering out duplicate values if wanted.
    with catch_warnings():
        filterwarnings("ignore", message="Observation names are not unique. To make them unique, call `.obs_names_make_unique`.")
        adata_combined = anndata.concat(
            {
                "DATA": adata_data,
                "REFERENCE": adata_ref
            },
            label=dge_comparison_column
        )

    if exclude_ambiguous_assignments:
        # check whether some cells are in both data and reference
        duplicated_mask = adata_combined.obs_names.duplicated(keep=False)

        if np.any(duplicated_mask):
            print("Exclude ambiguously assigned cells...")
            # remove duplicated values
            adata_combined = adata_combined[~duplicated_mask].copy()

    # add column to .obs for its use in rank_genes_groups()
    #adata_combined.obs = adata_combined.obs.filter([dge_comparison_column]) # empty obs

    print(f"Calculate differentially expressed genes with Scanpy's `rank_genes_groups` using '{method}'.")
    sc.tl.rank_genes_groups(adata=adata_combined,
                            groupby=dge_comparison_column,
                            groups=["DATA"],
                            reference="REFERENCE",
                            method=method,
                            )

    # create dataframe from results
    res_dict = create_deg_dataframe(
        adata=adata_combined, groups="DATA")
    df = res_dict["DATA"]

    if plot_volcano:
        cell_counts = adata_combined.obs[dge_comparison_column].value_counts()
        data_counts = cell_counts["DATA"]
        ref_counts = cell_counts["REFERENCE"]

        n_upreg = np.sum((df["pvals"] <= significance_threshold) & (df["logfoldchanges"] > fold_change_threshold))
        n_downreg = np.sum((df["pvals"] <= significance_threshold) & (df["logfoldchanges"] < -fold_change_threshold))

        config_table = pd.DataFrame({
            "": ["Annotation", "Cell type", "Region", "Cell number", "DEG number"],
            "Target": [elem[1] if isinstance(elem, tuple) else elem for elem in [target_annotation_tuple, target_cell_type_tuple, target_region_tuple]] + [data_counts, n_upreg],
            "Reference": [elem[1] if isinstance(elem, tuple) else elem for elem in [orig_ref_annotation_tuple, orig_ref_cell_type_tuple, ref_region_tuple]] + [ref_counts, n_downreg]
        })

        # remove empty rows
        config_table = config_table.set_index("").dropna(how="all").reset_index()

        volcano_plot(
            data=df,
            significance_threshold=significance_threshold,
            fold_change_threshold=fold_change_threshold,
            title=title,
            savepath = savepath,
            save_only = save_only,
            dpi_save = dpi_save,
            config_table = config_table,
            adjust_labels=True,
            **volcano_kwargs
            )
    if return_results:
        return {
            "results": df,
            "params": adata_combined.uns["rank_genes_groups"]["params"]
        }



import string
from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt

from insitupy.palettes import CustomPalettes

# make sure that images do not exceed limits in c++ (required for cv2::remap function in cv2::warpAffine)
# see also https://www.geeksforgeeks.org/climits-limits-h-cc/
SHRT_MAX = 2**15-1 # 32767
SHRT_MIN = -(2**15-1) # -32767

# create cache dir
CACHE = Path.home() / ".cache/InSituPy/"

# modalities
MODALITIES = ["annotations", "cells", "images", "regions", "transcripts"]
LOAD_FUNCS = [
    'load_annotations',
    'load_cells',
    'load_images',
    'load_regions',
    'load_transcripts'
    ]

# naming
ISPY_METADATA_FILE = ".ispy"
XENIUM_HEX_RANGE = string.ascii_lowercase[:16]
NORMAL_HEX_RANGE = "".join([str(e) for e in range(10)]) + string.ascii_lowercase[:6]
XENIUM_INT_TO_HEX_CONV_DICT = {k:v for k,v in zip(NORMAL_HEX_RANGE, XENIUM_HEX_RANGE)}
XENIUM_HEX_TO_INT_CONV_DICT = {v:k for k,v in zip(NORMAL_HEX_RANGE, XENIUM_HEX_RANGE)}

# napari layer symbols
# SHAPES_SYMBOL = "\u2605" # Star: ★
# POINTS_SYMBOL = "\u2022" # Bullet: •
ANNOTATIONS_SYMBOL = "\U0001F52C" # 🔬
POINTS_SYMBOL = "\U0001F4CD" # 📍
REGIONS_SYMBOL = "\U0001F30D" # 🌍

# annotations
FORBIDDEN_ANNOTATION_NAMES = ["rest"]

## Matplotlib settings
# cmaps
palettes = CustomPalettes()
DEFAULT_CATEGORICAL_CMAP = palettes.tab20_mod
REGION_CMAP = matplotlib.colormaps["tab10"]
DEFAULT_CONTINUOUS_CMAP = "viridis"

# font size
def _init_mpl_fontsize(scale_factor=1):
    '''
    https://matplotlib.org/stable/api/matplotlib_configuration_api.html#matplotlib.rcParams
    '''
    SMALL_SIZE = 14*scale_factor
    MEDIUM_SIZE = 16*scale_factor
    BIGGER_SIZE = 18*scale_factor

    plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
    plt.rc('axes', titlesize=MEDIUM_SIZE)     # fontsize of the axes title
    plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
    plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
    plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
    plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
    plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

_init_mpl_fontsize()


import dask
import numpy as np
import pandas as pd
from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.figure import Figure
from scipy.sparse import issparse

from insitupy import WITH_NAPARI

if WITH_NAPARI:
    def init_data_name():
        global current_data_name
        current_data_name = "main"

    def init_recent_selections():
        global recent_selections
        recent_selections = []

    def init_colorlegend_canvas():
        # set up colorlegend
        global static_canvas
        static_canvas = FigureCanvas(Figure(figsize=(5, 5)))

    # set viewer configurations
    def init_viewer_config(
        xdata,
        pixel_size_param = None
        ):
        if current_data_name == "main":
            # access adata, viewer and metadata from InSituData
            global adata
            adata = xdata.cells.matrix
            global boundaries
            boundaries = xdata.cells.boundaries
            global viewer
            viewer = xdata.viewer
        else:
            adata = xdata.alt[current_data_name].matrix
            boundaries = xdata.alt[current_data_name].boundaries

        # get keys from var_names, obs and obsm
        global genes, observations, value_dict
        genes = sorted(adata.var_names.tolist())
        observations = sorted(adata.obs.columns.tolist())

        obsm_keys = list(adata.obsm.keys())
        obsm_cats = []
        for k in sorted(obsm_keys):
            data = adata.obsm[k]
            if isinstance(data, pd.DataFrame):
                for col in data.columns:
                    obsm_cats.append(f"{k}#{col}")
            elif isinstance(data, np.ndarray):
                for i in range(data.shape[1]):
                    obsm_cats.append(f"{k}#{i+1}")
            else:
                pass

        value_dict = {
            "genes": genes,
            "obs": observations,
            "obsm": obsm_cats
        }

        # get point coordinates
        global points
        points = np.flip(adata.obsm["spatial"].copy(), axis=1) # switch x and y (napari uses [row,column])

        # get expression matrix
        global X
        if issparse(adata.X):
            X = adata.X.toarray()
        else:
            X = adata.X

        global masks
        masks = []
        for n in boundaries.metadata.keys():
            b = boundaries[n]
            if isinstance(b, dask.array.core.Array) or np.all([isinstance(elem, dask.array.core.Array) for elem in b]):
                masks.append(n)

        if xdata.images is not None:
            # get image metadata
            global pixel_size
            if pixel_size_param is None:
                first_key = list(xdata.images.metadata.keys())[0]
                pixel_size = xdata.images.metadata[first_key]["pixel_size"]
            else:
                pixel_size = pixel_size_param




import os
import warnings
from pathlib import Path
from typing import Union

import geopandas
import numpy as np
import pandas as pd
from geopandas.geodataframe import GeoDataFrame

from ..utils.utils import convert_to_list

# force geopandas to use shapely. Default in future versions of geopandas.
os.environ['USE_PYGEOS'] = '0'


def parse_geopandas(
    data: Union[GeoDataFrame, pd.DataFrame, dict,
                str, os.PathLike, Path],
    uid_col: str = "id"
    ):
    # check if the input is a path or a GeoDataFrame
    if isinstance(data, GeoDataFrame):
        df = data
        df["origin"] = "manual"
    elif isinstance(data, pd.DataFrame) or isinstance(data, dict):
        df = GeoDataFrame(data, geometry=data["geometry"])
        df["origin"] = "manual"
    else:
        # read annotations as GeoDataFrame
        data = Path(data)
        if data.suffix == ".geojson":
            df = read_qupath_geojson(file=data)
            df["origin"] = "file"
        else:
            raise ValueError(f"Unknown file extension: {data.suffix}. File is expected to be `.geojson` or `.parquet`.")

    # set the crs to EPSG:4326 (does not matter for us but to circumvent errors it is better to set it)
    df = df.set_crs(4326)

    if df.index.name != uid_col:
        # set uid column as index
        df = df.set_index(uid_col)

    return df

def read_qupath_geojson(file: Union[str, os.PathLike, Path]) -> pd.DataFrame:
    """
    Reads a QuPath-compatible GeoJSON file and transforms it into a flat DataFrame.

    Parameters:
    - file (Union[str, os.PathLike, Path]): The file path (as a string or pathlib.Path) of the QuPath GeoJSON file.

    Returns:
    pandas.DataFrame: A DataFrame with flattened columns including "name" and "color" extracted from the "classification" column.
    """
    # Read the GeoJSON file into a GeoDataFrame
    dataframe = geopandas.read_file(file)

    # annotation geojsons contain a classification column where each entry is a dict with name and color of the annotation
    if "classification" in dataframe.columns:
        # Flatten the "classification" column into separate "name" and "color" columns
        if "name" in dataframe.columns:
            warnings.warn(
                (
                    f"The geometries contain both a 'name' (set e.g. by 'Set properties' in QuPath) and a 'classification name'.\n"
                    f"Currently, the `read_qupath_geojson` function overwrites the name with the classification name and saves it in a column named just 'name'.\n"
                    f"This behavior might change in the future."
                )
                )
        try:
            dataframe["name"] = [elem["name"] if pd.notnull(elem) else "unclassified" for elem in dataframe["classification"]]
        except KeyError:
            pass
        try:
            dataframe["color"] = [elem["color"] if pd.notnull(elem) else [0,0,0] for elem in dataframe["classification"]]
        except KeyError:
            pass

        try:
            dataframe["scale"] = [elem["scale"] if pd.notnull(elem) else (1,1) for elem in dataframe["classification"]]
        except KeyError:
            pass

        # Remove the redundant "classification" column
        dataframe = dataframe.drop(["classification"], axis=1)

    # Exported TMA cores instead contain the columns 'name' and 'isMissing'. These we just leave.

    # Return the transformed DataFrame
    return dataframe

def write_qupath_geojson(dataframe: GeoDataFrame,
                         file: Union[str, os.PathLike, Path]
                         ):
    """
    Converts a GeoDataFrame with "name" and "color" columns into a QuPath-compatible GeoJSON-like format,
    adding a new "classification" column containing dictionaries with "name" and "color" entries.
    The modified GeoDataFrame is then saved to the specified GeoJSON file.

    Parameters:
    - dataframe (geopandas.GeoDataFrame): The input GeoDataFrame containing "name" and "color" columns.
    - file (Union[str, os.PathLike, Path]): The file path (as a string or pathlib.Path) where the GeoJSON data will be saved.
    """
    columns_to_move = ["name", "color", "scale"]
    if np.any([elem in dataframe.columns for elem in columns_to_move]):
        existing_columns_to_move = [elem for elem in columns_to_move if elem in dataframe.columns]

        # Initialize an empty list to store dictionaries for each row
        classification_list = []

        # Iterate over rows in the GeoDataFrame
        for _, row in dataframe.iterrows():
            # Create a dictionary with "name" and "color" entries for each row
            classification_dict = {}

            for column in existing_columns_to_move:
                entry = row[column]

                # convert numpy arrays to lists
                if isinstance(entry, np.ndarray):
                    entry = convert_to_list(entry)

                classification_dict[column] = entry
            # Append the dictionary to the list
            classification_list.append(classification_dict)

        # Add a new "classification" column to the GeoDataFrame
        dataframe["classification"] = classification_list

        # Remove the original "name" and "color" columns
        dataframe = dataframe.drop(existing_columns_to_move, axis=1)

    # try:
    #     # remove scale column
    #     dataframe = dataframe.drop("scale", axis=1)
    # except KeyError:
    #     pass

    # Write the GeoDataFrame to a GeoJSON file
    dataframe.to_file(file, driver="GeoJSON")




import hashlib
import os
import shutil
from pathlib import Path

from insitupy._constants import CACHE
from insitupy._core.insitudata import InSituData
from insitupy._core.reader import read_xenium
from insitupy.datasets.download import download_url

# parameters for download functions
DEMODIR = CACHE / 'demo_datasets'

# functions that each download a dataset into  '~/.cache/InSituPy/demo_dataset'
def md5sum(filePath):
    with open(filePath, 'rb') as fh:
        m = hashlib.md5()
        while True:
            data = fh.read(8192)
            if not data:
                break
            m.update(data)
    return m.hexdigest()

# function that checks the md5sum of the images and returns a boolean.
def md5sum_image_check(file_path : Path, expected_md5sum, overwrite):
    download = False
    if file_path.exists():
        print("Image exists. Checking md5sum...")
        if md5sum(file_path) == expected_md5sum:
            if not overwrite:
                print(f"The md5sum matches. Download is skipped. To force download set `overwrite=True`.")
                return
            else:
                # if the md5sum matches but overwrite=True, the image is still downloaded
                download = True
        else:
            print(f"The md5sum doesn't match. Image is downloaded.")
            download = True
    else:
        download = True

    return download


def list_downloaded_datasets():
    try:
        # List all items in the given directory
        items = os.listdir(DEMODIR)
        # Filter out only the folders
        folders = [item for item in items if os.path.isdir(os.path.join(DEMODIR, item))]

        if folders:
            print("Following demo datasets were found:\n")
            for folder in folders:
                print(f"- {folder}")
        else:
            print("No folders found in the specified directory.")
    except FileNotFoundError:
        print(f"The directory '{DEMODIR}' does not exist.")
    except PermissionError:
        print(f"Permission denied to access the directory '{DEMODIR}'.")


# function that checks data for md5sum, downloads and unpacks the data.
def data_check_and_download(data_dir, zip_file, expected_md5sum, overwrite, data_url, named_data_dir):
    # check if the unzipped data exists
    download_data = False
    if data_dir.exists():
        # if it exists, everything is fine and it is assumed that the dataset was downloaded correctly. Overwrite is still checked.
        if not overwrite:
            print(f"This dataset exists already. Download is skipped. To force download set `overwrite=True`.")
            return
        else:
            print(f"This dataset exists already but is overwritten because of `overwrite=True`.")
            download_data = True
    else:
        # if unzipped data does not exist, we need to check if a zip file exists, and if yes if its md5sum is correct
        if zip_file.exists():
            print("ZIP file exists. Checking md5sum...")
            if md5sum(zip_file) == expected_md5sum:
                if not overwrite:
                    print(f"This dataset exists already. Download is skipped. To force download set `overwrite=True`.")
                    return
                else:
                    # if the md5sum matches but overwrite=True, the data is still downloaded
                    download_data = True
            else:
                print("The dataset exists already but the md5sum is not as expected. Dataset is downloaded again.")
                download_data = True
        else:
            download_data = True

    if download_data:
        # download data as zip file
        download_url(data_url, out_dir=named_data_dir, overwrite=True)

        # unzip data
        shutil.unpack_archive(zip_file, data_dir)

        # move files from outs folder into parent directory and remove the outs folder
        if (data_dir / "outs/").exists():
            for f in data_dir.glob("outs/*"):
                shutil.move(f, data_dir)
            os.rmdir(data_dir / "outs")

        #remove zip file after unpacking
        os.remove(zip_file)


# Xenium onboard analysis version 1.0.1
# data from https://www.10xgenomics.com/products/xenium-in-situ/preview-dataset-human-breast
def human_breast_cancer(
        overwrite: bool = False
) -> InSituData:

    # URLs for download
    data_url = "https://cf.10xgenomics.com/samples/xenium/1.0.1/Xenium_FFPE_Human_Breast_Cancer_Rep1/Xenium_FFPE_Human_Breast_Cancer_Rep1_outs.zip"
    he_url = "https://cf.10xgenomics.com/samples/xenium/1.0.1/Xenium_FFPE_Human_Breast_Cancer_Rep1/Xenium_FFPE_Human_Breast_Cancer_Rep1_he_image.ome.tif"
    if_url = "https://cf.10xgenomics.com/samples/xenium/1.0.1/Xenium_FFPE_Human_Breast_Cancer_Rep1/Xenium_FFPE_Human_Breast_Cancer_Rep1_if_image.ome.tif"

    # set up paths
    named_data_dir = DEMODIR / "hbreastcancer"
    data_dir = named_data_dir / "output-XETG00000__slide_id__hbreastcancer"
    image_dir = named_data_dir / "unregistered_images"
    zip_file = named_data_dir / Path(data_url).name

    # check if file exists and has correct md5sum
    expected_md5sum = '7d42a0b232f92a2e51de1f513b1a44fd'
    expected_he_md5sum = 'fc0d0d38b7c039cc0682e51099f8d841'
    expected_if_md5sum = '929839c64ef8331cfd048a614f5f6829'

    # image file names
    he_file_name = "slide_id__hbreastcancer__HE__histo"
    if_file_name = "slide_id__hbreastcancer__CD20_HER2_DAPI__IF"
    # check if data exists (zipped or unzipped), if yes check md5sum
    # if necessary download data
    data_check_and_download(data_dir, zip_file, expected_md5sum, overwrite, data_url, named_data_dir)

    # download image data
    if md5sum_image_check(image_dir/"slide_id__hbreastcancer__HE__histo.ome.tif", expected_he_md5sum, overwrite):
        download_url(he_url, out_dir = image_dir, file_name = he_file_name, overwrite = True)

    if md5sum_image_check(image_dir/"slide_id__hbreastcancer__CD20_HER2_DAPI__IF.ome.tif", expected_if_md5sum, overwrite):
        download_url(if_url, out_dir = image_dir, file_name = if_file_name, overwrite = True)

    print(f"Corresponding image data can be found in {image_dir}")
    print("For this dataset following images are available:")
    print(f"{he_file_name}.ome.tiff")
    print(f"{if_file_name}.ome.tiff")

    # load data into InSituData object
    data = read_xenium(data_dir)

    return data

# xenium onboard analysis version 1.5.0
# data from https://www.10xgenomics.com/resources/datasets/human-kidney-preview-data-xenium-human-multi-tissue-and-cancer-panel-1-standard
def human_kidney_nondiseased(
        overwrite: bool = False
) -> InSituData:

    # URLs for download
    data_url = "https://cf.10xgenomics.com/samples/xenium/1.5.0/Xenium_V1_hKidney_nondiseased_section/Xenium_V1_hKidney_nondiseased_section_outs.zip"
    he_url = "https://cf.10xgenomics.com/samples/xenium/1.5.0/Xenium_V1_hKidney_nondiseased_section/Xenium_V1_hKidney_nondiseased_section_he_image.ome.tif"

    # set up paths
    named_data_dir = DEMODIR / "hkidney"
    data_dir = named_data_dir / "output-XETG0000__slide_id__hkidney"
    image_dir = named_data_dir / "unregistered_images"
    zip_file = named_data_dir / Path(data_url).name

    # check if file exists and has correct md5sum
    expected_md5sum = '194d5e21b40b27fa8c009d4cbdc3272d'
    expected_he_md5sum = 'e457889aea78bef43834e675f0c58d95'

    # image file name
    he_file_name = "slide_id__hkidney__HE__histo"

    # check if data exists (zipped or unzipped), if yes check md5sum
    # if necessary download data
    data_check_and_download(data_dir, zip_file, expected_md5sum, overwrite, data_url, named_data_dir)

    # download image data
    if md5sum_image_check(image_dir/"slide_id__hkidney__HE__histo.ome.tif", expected_he_md5sum, overwrite):
        download_url(he_url, out_dir = image_dir, file_name = he_file_name, overwrite = True)

    print(f"Corresponding image data can be found in {image_dir}")
    print("For this dataset following image is available:")
    print(f"{he_file_name}.ome.tiff")

    # load data into InSituData object
    data = read_xenium(data_dir)

    return data

# xenium onboard analysis version 1.6.0
# data from https://www.10xgenomics.com/datasets/pancreatic-cancer-with-xenium-human-multi-tissue-and-cancer-panel-1-standard
def human_pancreatic_cancer(
        overwrite: bool = False

) -> InSituData:

    # URLs for download
    data_url = "https://cf.10xgenomics.com/samples/xenium/1.6.0/Xenium_V1_hPancreas_Cancer_Add_on_FFPE/Xenium_V1_hPancreas_Cancer_Add_on_FFPE_outs.zip"
    he_url = "https://cf.10xgenomics.com/samples/xenium/1.6.0/Xenium_V1_hPancreas_Cancer_Add_on_FFPE/Xenium_V1_hPancreas_Cancer_Add_on_FFPE_he_image.ome.tif"
    if_url = "https://cf.10xgenomics.com/samples/xenium/1.6.0/Xenium_V1_hPancreas_Cancer_Add_on_FFPE/Xenium_V1_hPancreas_Cancer_Add_on_FFPE_if_image.ome.tif"

    # set up paths
    named_data_dir = DEMODIR / "hpancreas"
    data_dir = named_data_dir / "output-XETG00000__slide_id__hpancreas"
    image_dir = named_data_dir / "unregistered_images"
    zip_file = named_data_dir / Path(data_url).name

    # check if file exists and has correct md5sum
    expected_md5sum = '7acca4c2a40f09968b72275403c29f93'
    expected_he_md5sum = '4e96596ea13a3d0f6139638b2b90aef4'
    expected_if_md5sum = 'c859a7ab5d29807b4daf1f66cb6f5060'

    # image file names
    he_file_name = "slide_id__hPancreas__HE__histo"
    if_file_name = "slide_id__hPancreas__CD20_TROP2_PPY_DAPI__IF"

    # check if data exists (zipped or unzipped), if yes check md5sum
    # if necessary download data)
    data_check_and_download(data_dir, zip_file, expected_md5sum, overwrite, data_url, named_data_dir)

    # download image data
    if md5sum_image_check(image_dir/"slide_id__hPancreas__HE__histo.ome.tif", expected_he_md5sum, overwrite):
        download_url(he_url, out_dir = image_dir, file_name = he_file_name, overwrite = True)

    if md5sum_image_check(image_dir/"slide_id__hPancreas__CD20_TROP2_PPY_DAPI__IF.ome.tif", expected_if_md5sum, overwrite):
        download_url(if_url, out_dir = image_dir, file_name = if_file_name, overwrite = True )

    print(f"Corresponding image data can be found in {image_dir}")
    print("For this dataset following images are available:")
    print(f"{he_file_name}.ome.tiff")
    print(f"{if_file_name}IF_image_name.ome.tiff")

    # load data into InSituData object
    data = read_xenium(data_dir)

    return data

# xenium onboard analysis version 1.7.0
# data from https://www.10xgenomics.com/resources/datasets/human-skin-preview-data-xenium-human-skin-gene-expression-panel-add-on-1-standard
def human_skin_melanoma(
        overwrite: bool = False

) -> InSituData:

    # URLs for download
    data_url = "https://cf.10xgenomics.com/samples/xenium/1.7.0/Xeniumranger_V1_hSkin_Melanoma_Add_on_FFPE/Xeniumranger_V1_hSkin_Melanoma_Add_on_FFPE_outs.zip"
    he_url = "https://cf.10xgenomics.com/samples/xenium/1.7.0/Xeniumranger_V1_hSkin_Melanoma_Add_on_FFPE/Xeniumranger_V1_hSkin_Melanoma_Add_on_FFPE_he_image.ome.tif"

    # set up paths
    named_data_dir = DEMODIR / "hskin"
    data_dir = named_data_dir / "output-XETG00000__slide_id__hskin"
    image_dir = named_data_dir / "unregistered_images"
    zip_file = named_data_dir / Path(data_url).name

    # check if file exists and has correct md5sum
    expected_md5sum = '29102799a3f1858c7318b705eb1a8584'
    expected_he_md5sum = '169af7630e0124eef61d252183243a06'

    # image file name
    he_file_name = "slide_id__hskin__HE__histo"

    # check if data exists (zipped or unzipped), if yes check md5sum
    # if necessary download data
    data_check_and_download(data_dir, zip_file, expected_md5sum, overwrite, data_url, named_data_dir)

    # download image data
    if md5sum_image_check(image_dir/"slide_id__hskin__HE__histo.ome.tif", expected_he_md5sum, overwrite):
        download_url(he_url, out_dir = image_dir, file_name = he_file_name, overwrite = True)

    print(f"Corresponding image data can be found in {image_dir}")
    print("For this dataset following image is available:")
    print(f"{he_file_name}.ome.tiff")

    # load data into InSituData object
    data = read_xenium(data_dir)

    return data

# xenium onboard analysis version 2.0.0
# data from https://www.10xgenomics.com/datasets/ffpe-human-brain-cancer-data-with-human-immuno-oncology-profiling-panel-and-custom-add-on-1-standard
def human_brain_cancer(
        overwrite: bool = False

) -> InSituData:

    # URLs for download
    data_url = "https://s3-us-west-2.amazonaws.com/10x.files/samples/xenium/2.0.0/Xenium_V1_Human_Brain_GBM_FFPE/Xenium_V1_Human_Brain_GBM_FFPE_outs.zip"
    he_url = "https://cf.10xgenomics.com/samples/xenium/2.0.0/Xenium_V1_Human_Brain_GBM_FFPE/Xenium_V1_Human_Brain_GBM_FFPE_he_image.ome.tif"

    # set up paths
    named_data_dir = DEMODIR / "hbraincancer"
    data_dir = named_data_dir / "output-XETG00000__slide_id__hbraincancer"
    image_dir = named_data_dir / "unregistered_images"
    zip_file = named_data_dir / Path(data_url).name

    # check if file exists and has correct md5sum
    expected_md5sum = "c116017ad9884cf6944c6c4815bffb3c"
    expected_he_md5sum = "22b66c6e7669933e50a9665d467e639f"

    # image file name
    he_file_name = "slide_id__hbraincancer__HE__histo"

    # check if data exists (zipped or unzipped), if yes check md5sum
    # if necessary download data
    data_check_and_download(data_dir, zip_file, expected_md5sum, overwrite, data_url, named_data_dir)

    # download image data
    if md5sum_image_check(image_dir/"slide_id__hbraincancer__HE__histo.ome.tif", expected_he_md5sum, overwrite):
        download_url(he_url, out_dir = image_dir, file_name = he_file_name, overwrite = True)

    print(f"Corresponding image data can be found in {image_dir}")
    print("For this dataset following image is available:")
    print(f"{he_file_name}.ome.tiff")

    # load data into InSituData object
    data = read_xenium(data_dir)

    return data

# xenium onboard analysis 2.0.0
# data from https://www.10xgenomics.com/datasets/preview-data-ffpe-human-lung-cancer-with-xenium-multimodal-cell-segmentation-1-standard
def human_lung_cancer(
        overwrite: bool = False

) -> InSituData:

    # URLs for download
    data_url = "https://cf.10xgenomics.com/samples/xenium/2.0.0/Xenium_V1_humanLung_Cancer_FFPE/Xenium_V1_humanLung_Cancer_FFPE_outs.zip"
    he_url = "https://cf.10xgenomics.com/samples/xenium/2.0.0/Xenium_V1_humanLung_Cancer_FFPE/Xenium_V1_humanLung_Cancer_FFPE_he_image.ome.tif"

    # set up paths
    named_data_dir = DEMODIR / "hlungcancer"
    data_dir = named_data_dir / "output-XETG00000__slide_id__hlungcancer"
    image_dir = named_data_dir / "unregistered_images"
    zip_file = named_data_dir / Path(data_url).name

    # check if file exists and has correct md5sum
    expected_md5sum = "194e24c1efe7e64d2487adfe313bb9dd"
    expected_he_md5sum = "47147933d73e008a0dd3695895832dd4"

    # image file names
    he_file_name = "slide_id__hlungcancer__HE__histo"

    # check if data exists (zipped or unzipped), if yes check md5sum
    # if necessary download data
    data_check_and_download(data_dir, zip_file, expected_md5sum, overwrite, data_url, named_data_dir)

    # download image data
    if md5sum_image_check(image_dir/"slide_id__hlungcancer__HE__histo.ome.tif", expected_he_md5sum, overwrite):
        download_url(he_url, out_dir = image_dir, file_name = he_file_name, overwrite = True)

    print(f"Corresponding image data can be found in {image_dir}")
    print("For this dataset following image is available:")
    print(f"{he_file_name}.ome.tiff")

    # load data into InSituData object
    data = read_xenium(data_dir)

    return data

# xenium onboard analysis 3.0.0
# data from https://www.10xgenomics.com/datasets/preview-data-xenium-prime-gene-expression
def human_lymph_node_5k(
        overwrite: bool = False

) -> InSituData:

    # URLs for download
    data_url = "https://s3-us-west-2.amazonaws.com/10x.files/samples/xenium/3.0.0/Xenium_Prime_Human_Lymph_Node_Reactive_FFPE/Xenium_Prime_Human_Lymph_Node_Reactive_FFPE_outs.zip"
    he_url = "https://cf.10xgenomics.com/samples/xenium/3.0.0/Xenium_Prime_Human_Lymph_Node_Reactive_FFPE/Xenium_Prime_Human_Lymph_Node_Reactive_FFPE_he_image.ome.tif"

    # set up paths
    named_data_dir = DEMODIR / "hlymphnode5k"
    data_dir = named_data_dir / "output-XETG00000__slide_id__hlymphnode5k"
    image_dir = named_data_dir / "unregistered_images"
    zip_file = named_data_dir / Path(data_url).name

    # check if file exists and has correct md5sum
    expected_md5sum = "1ddb7d10e2bca93a61830d5dc57cb3a8"
    expected_he_md5sum = "eae078b0b3ddcaf6a11ca4eefabcaf0c"

    # image file names
    he_file_name = "slide_id__hlymphnode5k__HE__histo"

    # check if data exists (zipped or unzipped), if yes check md5sum
    # if necessary download data
    data_check_and_download(data_dir, zip_file, expected_md5sum, overwrite, data_url, named_data_dir)

    # download image data
    if md5sum_image_check(image_dir/"slide_id__hlymphnode5k__HE__histo.ome.tif", expected_he_md5sum, overwrite):
        download_url(he_url, out_dir = image_dir, file_name = he_file_name, overwrite = True)

    print(f"Corresponding image data can be found in {image_dir}")
    print("For this dataset following image is available:")
    print(f"{he_file_name}.ome.tiff")

    # load data into InSituData object
    data = read_xenium(data_dir)

    return data

# xenium onboard analysis 1.5.0
# data from https://www.10xgenomics.com/datasets/human-lymph-node-preview-data-xenium-human-multi-tissue-and-cancer-panel-1-standard
def human_lymph_node(
        overwrite: bool = False

) -> InSituData:

    # URLs for download
    data_url = "https://cf.10xgenomics.com/samples/xenium/1.5.0/Xenium_V1_hLymphNode_nondiseased_section/Xenium_V1_hLymphNode_nondiseased_section_outs.zip"

    # set up paths
    named_data_dir = DEMODIR / "hlymphnode"
    data_dir = named_data_dir / "output-XETG00000__slide_id__hlymphnode"
    image_dir = named_data_dir / "unregistered_images"
    zip_file = named_data_dir / Path(data_url).name

    # check if file exists and has correct md5sum
    expected_md5sum = "4e9ef0a40b0fc00e619cf310f349f1bd"

    # check if data exists (zipped or unzipped), if yes check md5sum
    # if necessary download data
    data_check_and_download(data_dir, zip_file, expected_md5sum, overwrite, data_url, named_data_dir)

    print('For this dataset no image is available')

    # load data into InSituData object
    data = read_xenium(data_dir)

    return data


# xenium onboard analysis 3.0.0
# data from https://cf.10xgenomics.com/samples/xenium/3.0.0/Xenium_Prime_Mouse_Ileum_tiny/Xenium_Prime_Mouse_Ileum_tiny_outs.zip
# Mouse ileum, nuclear expansion
def xenium_test_dataset(
        overwrite: bool = False
) -> InSituData:
    # URLs for download
    data_url = "https://cf.10xgenomics.com/samples/xenium/3.0.0/Xenium_Prime_Mouse_Ileum_tiny/Xenium_Prime_Mouse_Ileum_tiny_outs.zip"

    # set up paths
    named_data_dir = DEMODIR / "xenium_test_dataset"
    data_dir = named_data_dir / "output-XETG00000__slide_id__xenium_test_dataset"
    image_dir = named_data_dir / "unregistered_images"
    zip_file = named_data_dir / Path(data_url).name


    # check if file exists and has correct md5sum
    expected_md5sum = "0a7469a005576f2932e4f804dd9bc563"

    # check if data exists (zipped or unzipped), if yes check md5sum
    # if necessary download data
    data_check_and_download(data_dir, zip_file, expected_md5sum, overwrite, data_url, named_data_dir)

    # load data into InSituData object
    data = read_xenium(data_dir)

    print('For this dataset no image is available')
    return data